// MgmtInfoBase.cpp : CMgmtInfoBase ̃Cve[V
#include "stdafx.h"
#include "SNMPCtl.h"
#include "MgmtInfoBase.h"

/////////////////////////////////////////////////////////////////////////////
// CMgmtInfoBase
oid RFC1213_MIB[] = { 1, 3, 6, 1, 2, 1 };
const unsigned char RFC1213_MIB_text[] = ".iso.org.dod.internet.mgmt.mib-2";
const unsigned char EXPERIMENTAL_MIB_text[] = ".iso.org.dod.internet.experimental";
const unsigned char PRIVATE_MIB_text[] = ".iso.org.dod.internet.private";
const unsigned char PARTY_MIB_text[] = ".iso.org.dod.internet.snmpParties";
const unsigned char SECRETS_MIB_text[] = ".iso.org.dod.internet.snmpSecrets";

static char Standard_Prefix[] = ".1.3.6.1.2.1.";

STDMETHODIMP CMgmtInfoBase::InterfaceSupportsErrorInfo(REFIID riid)
{
	static const IID* arr[] = 
	{
		&IID_IMgmtInfoBase
	};
	for (int i=0; i < sizeof(arr) / sizeof(arr[0]); i++)
	{
		if (::InlineIsEqualGUID(*arr[i],riid))
			return S_OK;
	}
	return S_FALSE;
}

CMgmtInfoBase::CMgmtInfoBase()
	: m_strPrefix(Standard_Prefix), m_mib(NULL)
{
	m_pUnkMarshaler = NULL;
}

void CMgmtInfoBase::ChangePrefix(std::string& str)
{
	Lock();
	if (m_strPrefix[0] == '.')
		str = m_strPrefix.c_str() + 1;
	else
		str = m_strPrefix;
	Unlock();
	if (str[str.length() - 1] != '.')
		str += '.';
}

HRESULT STDMETHODCALLTYPE CMgmtInfoBase::put_Prefix( 
            /* [in] */ BSTR NewVal)
{
	if (!NewVal) NewVal = L"";
	USES_CONVERSION;
	Lock();
	m_strPrefix = W2A(NewVal);
	Unlock();
	return S_OK;
}
        
HRESULT STDMETHODCALLTYPE CMgmtInfoBase::get_Prefix( 
			 /* [retval][out] */ BSTR __RPC_FAR *pRetVal)
{
	if (!pRetVal) return E_POINTER;
	USES_CONVERSION;
	Lock();
	*pRetVal = SysAllocString(A2W(m_strPrefix.c_str()));
	Unlock();
	return S_OK;
}
        
HRESULT STDMETHODCALLTYPE CMgmtInfoBase::InitNew( void)
{
	HRESULT hr = S_OK;
	Lock();
	if (m_mib)
		CMIB::DestroyTree(m_mib);
	// ݁AftHgȊOMIB.TXT󂯕tȂB
	// IStreamC^[tFCXŎ󂯕t悤ɕύX\
	HRSRC hrsrc = FindResource(_Module.GetResourceInstance(), MAKEINTRESOURCE(IDR_MIBTEXT), _T("TEXT"));
	_ASSERT(hrsrc);
	HGLOBAL h = LoadResource(_Module.GetResourceInstance(), hrsrc);
	_ASSERT(h);
	LPSTR lp = (LPSTR)LockResource(h);
	_ASSERT(lp);
	DWORD dw = SizeofResource(_Module.GetResourceInstance(), hrsrc);
	if (*(lp + dw))	// If there is no null terminator, then, copy into the stack.
	{
		LPSTR p = (LPSTR)_alloca(dw + 1);
		CopyMemory(p, lp, dw);
		*(lp + dw) = '\0';
		lp = p;
	}
	try
	{
		m_mib = CMIB::CreateTree(lp);
	}
	catch (std::string* p)
	{
		hr = Error(p->c_str());
		delete p;
	}
	Unlock();
	return hr;
}

HRESULT STDMETHODCALLTYPE CMgmtInfoBase::ReadObjId( 
            /* [in] */ BSTR ObjectString,
            /* [retval][out] */ VARIANT __RPC_FAR *pResult)
{
	if (!pResult || !ObjectString) return E_POINTER;
    if (m_mib == NULL)
		return DISP_E_EXCEPTION;

    char buf[512];

	USES_CONVERSION;
	char* input = W2A(ObjectString);
	int len = lstrlenA(input);
	if (len >= sizeof(buf)) 
		return DISP_E_OVERFLOW;
    if (*input == '.')
		input++;
    else {
		std::string s;
		ChangePrefix(s);
		if (len >= sizeof(buf) - s.length())
			return DISP_E_OVERFLOW;
        lstrcpyA(buf, s.c_str());
		lstrcatA(buf, input);
		input = buf;
    }

	std::vector<oid> output;
	try
	{
		m_mib->ParseSubTree(input, output);
	}
	catch (std::string* p)
	{
		Error(p->c_str());
		delete p;
		return DISP_E_EXCEPTION;
	}

	if (output.size() == 0)
	{
		std::wstring s = L"Invalid object identifier:";
		s += ObjectString;
		return Error(s.c_str());
	}
	else
	{
		SAFEARRAYBOUND sab[1];
		sab[0].lLbound = 0;
		sab[0].cElements = output.size();
		SAFEARRAY* pAry = SafeArrayCreate(VT_VARIANT, 1, sab);
		SafeArrayLock(pAry);
		VARIANT* p = (VARIANT*)pAry->pvData;
		for (int i = 0; i < output.size(); i++, p++)
		{
			VariantInit(p);
			p->vt = VT_I4;
			p->lVal = output[i];
		}
		SafeArrayUnlock(pAry);
		VariantInit(pResult);
		pResult->vt = VT_VARIANT | VT_ARRAY;
		pResult->parray = pAry;
	}
	return S_OK;
}
        
HRESULT STDMETHODCALLTYPE CMgmtInfoBase::PrintObjId( 
            /* [in] */ VARIANT ObjectId,
            /* [retval][out] */ BSTR __RPC_FAR *pResult)
{
	if (!pResult) return E_POINTER;
    if (m_mib == NULL)
		return DISP_E_EXCEPTION;
	SAFEARRAY* pAry = NULL;
	if ((ObjectId.vt & (VT_ARRAY | VT_BYREF)) == (VT_ARRAY | VT_BYREF))
		pAry = *ObjectId.pparray;
	else if (ObjectId.vt & VT_ARRAY)
		pAry = ObjectId.parray;
	else
		return Error("ObjectId must be an array.", GUID_NULL, DISP_E_BADVARTYPE);
	if (pAry->rgsabound[0].cElements == 0)
		return Error("No ObjectId", GUID_NULL, DISP_E_PARAMNOTFOUND);
	int vt = ObjectId.vt & ~(VT_ARRAY | VT_BYREF);
	VARIANT* pVar = NULL;
	long* pLong = NULL;
	short* pShort = NULL;
	LPBYTE pByte = NULL;
	if (vt == VT_I4)
		pLong = (long*)pAry->pvData;
	else if (vt == VT_I2)
		pShort = (short*)pAry->pvData;
	else if (vt == VT_VARIANT)
		pVar = (VARIANT*)pAry->pvData;
	else if (vt == VT_UI1)
		pByte = (LPBYTE)pAry->pvData;
	else
		return Error("ObjectId elements must be integers.", GUID_NULL, DISP_E_BADVARTYPE);
	oid* oids = (oid*)_alloca(pAry->rgsabound[0].cElements * sizeof(oid));
	for (int i = 0; i < pAry->rgsabound[0].cElements; i++)
	{
		if (pLong)
			oids[i] = *pLong++;
		else if (pShort)
			oids[i] = *pShort++;
		else if (pByte)
			oids[i] = *pByte++;
		else
		{
			VARIANT v;
			VariantInit(&v);
			HRESULT hr = VariantChangeType(&v, pVar, VARIANT_NOVALUEPROP, VT_I4);
			if (hr != S_OK)
				return Error("ObjectId elements must be integers.", GUID_NULL, hr);
			oids[i] = v.lVal;
			pVar++;
		}
	}
	std::string sResult(".");
	CMIB::GetSymbol(oids, pAry->rgsabound[0].cElements, m_mib, sResult);
	USES_CONVERSION;
	*pResult = SysAllocString(A2W(sResult.c_str()));
	return S_OK;
}

