// SNMPClient.cpp : CSNMPClient ̃Cve[V

#include "stdafx.h"
#include "SNMPCtl.h"
#include "SNMPClient.h"
#include "SNMPPDU.h"

CComMultiThreadModel::AutoCriticalSection CSNMPClient::s_crit;
int CSNMPClient::g_Count(0);
/////////////////////////////////////////////////////////////////////////////
// CSNMPClient

HRESULT CSNMPClient::FinalConstruct()
{
	s_crit.Lock();
	g_Count++;
	if (g_Count == 1)
	{
		WSAData wsadata;
		WSAStartup(MAKEWORD(2, 0), &wsadata);
	}
	s_crit.Unlock();
	return S_OK;
}

void CSNMPClient::FinalRelease()
{
	if (m_sockThread)
	{
		::PostThreadMessage(m_idThread, WM_QUIT, 0, 0);
		::Sleep(500);
		::CloseHandle(m_sockThread);
	}
	if (m_pMgmtInfoBase)
		m_pMgmtInfoBase->Release();
	if (m_sock != INVALID_SOCKET)
		closesocket(m_sock);
	s_crit.Lock();
	g_Count--;
	if (g_Count == 0)
	{
		WSACleanup();
	}
	s_crit.Unlock();
}

HRESULT STDMETHODCALLTYPE CSNMPClient::put_PeerName( 
            /* [in] */ BSTR NewVal)
{
	if (!NewVal || !*NewVal) return E_POINTER;
	if (m_sock != INVALID_SOCKET) return S_OK;

	USES_CONVERSION;
	char* p = W2A(NewVal);
	if (m_strPeerName != p)
		m_strPeerName = p;
	return S_OK;
}
        
HRESULT STDMETHODCALLTYPE CSNMPClient::get_PeerName( 
            /* [retval][out] */ BSTR __RPC_FAR *pRetVal)
{
	if (!pRetVal) return E_POINTER;

	USES_CONVERSION;
	*pRetVal = SysAllocString(A2W(m_strPeerName.c_str()));
	return S_OK;
}
        
HRESULT STDMETHODCALLTYPE CSNMPClient::put_Port( 
            /* [in] */ short NewVal)
{
	if (m_sock != INVALID_SOCKET) return S_OK;
	if (m_sPort != NewVal)
		m_sPort = NewVal;
	return S_OK;
}
        
HRESULT STDMETHODCALLTYPE CSNMPClient::get_Port( 
            /* [retval][out] */ short __RPC_FAR *pRetVal)
{
	if (!pRetVal) return E_POINTER;
	*pRetVal = m_sPort;
	return S_OK;
}
        
HRESULT STDMETHODCALLTYPE CSNMPClient::put_Community( 
            /* [in] */ BSTR NewVal)
{
	if (m_sock != INVALID_SOCKET) return S_OK;

	USES_CONVERSION;
	m_strCommunity = (!NewVal) ? "" : W2A(NewVal);
	return S_OK;
}
        
HRESULT STDMETHODCALLTYPE CSNMPClient::get_Community( 
            /* [retval][out] */ BSTR __RPC_FAR *pRetVal)
{
	if (!pRetVal) return E_POINTER;

	USES_CONVERSION;
	*pRetVal = SysAllocString(A2W(m_strCommunity.c_str()));
	return S_OK;
}
        
HRESULT STDMETHODCALLTYPE CSNMPClient::put_Retries( 
            /* [in] */ short NewVal)
{
	if (NewVal == SNMP_DEFAULT_RETRIES)
		m_sRetries = DEFAULT_RETRIES;
	else
		m_sRetries = NewVal;
	return S_OK;
}
        
HRESULT STDMETHODCALLTYPE CSNMPClient::get_Retries( 
            /* [retval][out] */ short __RPC_FAR *pRetVal)
{
	if (!pRetVal) return E_POINTER;

	*pRetVal = m_sRetries;
	return S_OK;
}
        
HRESULT STDMETHODCALLTYPE CSNMPClient::put_Timeout( 
            /* [in] */ long NewVal)
{
	if (NewVal == SNMP_DEFAULT_TIMEOUT)
		m_lTimeout = DEFAULT_TIMEOUT;
	else
		m_lTimeout = NewVal;
	return S_OK;
}
        
HRESULT STDMETHODCALLTYPE CSNMPClient::get_Timeout( 
            /* [retval][out] */ long __RPC_FAR *pRetVal)
{
	if (!pRetVal) return E_POINTER;

	*pRetVal = m_lTimeout;
	return S_OK;
}

HRESULT STDMETHODCALLTYPE CSNMPClient::put_Version( 
            /* [in] */ short NewVal)
{
	if (NewVal < 0 || NewVal > SNMP_VERSION_2) return Error(IDS_E_VERSION);
	m_nVersion = NewVal;
	return S_OK;
}
        
HRESULT STDMETHODCALLTYPE CSNMPClient::get_Version( 
            /* [retval][out] */ short __RPC_FAR *pRetVal)
{
	if (!pRetVal) return E_POINTER;

	*pRetVal = (short)m_nVersion;
	return S_OK;
}

HRESULT STDMETHODCALLTYPE CSNMPClient::put_MgmtInfoBase( 
            /* [in] */ IMgmtInfoBase __RPC_FAR *pNewVal)
{
	if (pNewVal == NULL) return E_POINTER;
	IMgmtInfoBase* pBase;
	HRESULT hr = pNewVal->QueryInterface(IID_IMgmtInfoBase, (void**)&pBase);
	if (hr == S_OK)
	{
		if (m_pMgmtInfoBase)
			m_pMgmtInfoBase->Release();
		m_pMgmtInfoBase = pBase;
	}
	return hr;
}
        
HRESULT STDMETHODCALLTYPE CSNMPClient::get_MgmtInfoBase( 
            /* [retval][out] */ IMgmtInfoBase __RPC_FAR *__RPC_FAR *ppRetVal)
{
	if (!ppRetVal) return E_POINTER;
	if (m_pMgmtInfoBase == NULL) return Error(IDS_E_NOMIB);
	return m_pMgmtInfoBase->QueryInterface(IID_IMgmtInfoBase, (void**)ppRetVal);
}

HRESULT STDMETHODCALLTYPE CSNMPClient::Open( void)
{
	if (m_sock != INVALID_SOCKET) return S_OK;
	m_sock = WSASocket(AF_INET, SOCK_DGRAM, 0, NULL, 0, WSA_FLAG_OVERLAPPED);
	if (m_sock == INVALID_SOCKET)
	{
		return Error(IDS_E_GENERR);
	}
	unsigned long ul = inet_addr(m_strPeerName.c_str());
	if (ul == INADDR_NONE)
	{
		struct hostent* pHost = gethostbyname(m_strPeerName.c_str());
		if (!pHost)
		{
			return Error(IDS_E_ADDRESS);
		}
		ul = *(unsigned long*)pHost->h_addr;
	}
	ZeroMemory(&m_addr, sizeof(sockaddr_in));
	m_addr.sin_family = AF_INET;
	m_addr.sin_port = htons(m_sPort);
	m_addr.sin_addr.S_un.S_addr = ul;

	sockaddr_in  me;
	ZeroMemory(&me, sizeof(sockaddr_in));
	me.sin_family = AF_INET;
	me.sin_addr.S_un.S_addr = INADDR_ANY;
	if (bind(m_sock, (sockaddr*)&me, sizeof(sockaddr_in)))
	{
		return Error(IDS_E_LOCPORT);
	}
	ATLTRACE(_T("My PORT#=%d\n"), ntohs(me.sin_port));
	if (m_strCommunity[0] == '+')
	{
		m_nVersion = SNMP_VERSION_2C;
		m_strCommunity.erase(0, 1);
	}
	if (m_strCommunity[0] == '/')
	{
		// Authentification required
		// -- not supported, now
		m_strCommunity.erase(0, 1);
	}

	return S_OK;
}
        
HRESULT STDMETHODCALLTYPE CSNMPClient::Close( void)
{
	if (m_sockThread)
	{
		::PostThreadMessage(m_idThread, WM_QUIT, 0, 0);
		::Sleep(500);
		::CloseHandle(m_sockThread);
		m_sockThread = NULL;
	}
	if (m_sock != INVALID_SOCKET)
	{
		closesocket(m_sock);
		m_sock = INVALID_SOCKET;
	}
	return S_OK;
}

HRESULT STDMETHODCALLTYPE CSNMPClient::NewPDU( 
            /* [retval][out] */ IDispatch __RPC_FAR *__RPC_FAR *ppResult)
{
	if (!ppResult) return E_POINTER;

	CComObject<CSNMPPDU>* p;
	HRESULT hr = CComObject<CSNMPPDU>::CreateInstance(&p);
	if (hr == S_OK)
		return p->QueryInterface(IID_IDispatch, (void**)ppResult);
		
	return hr;
}
        
HRESULT STDMETHODCALLTYPE CSNMPClient::Request( 
            /* [in] */ IDispatch __RPC_FAR *pPDU,
            /* [retval][out] */ long __RPC_FAR *pReqid)
{
	if (!pPDU || !pReqid) return E_POINTER;
	if (m_sock == INVALID_SOCKET) return Error(IDS_E_GENERR);
	*pReqid = -1;
	CComQIPtr<ISNMPPDU, &IID_ISNMPPDU> p(pPDU);
	if (p == NULL) return Error(IDS_E_ARG);

	if (m_sockThread)
	{
		DWORD ExitCode;
		if (GetExitCodeThread(m_sockThread, &ExitCode))
		{
			if (ExitCode != STILL_ACTIVE)
			{
				CloseHandle(m_sockThread);
				m_sockThread = NULL;
			}
		}
  	}
	if (m_sockThread == NULL)
	{
		ThreadParam* pParam = new ThreadParam;
		pParam->sock = m_sock;
		pParam->addr = &m_addr;
		pParam->timeout = m_lTimeout;
		if (m_hWndCD)
			pParam->hWnd = m_hWndCD;
		else
		{
			pParam->hWnd = NULL;
			CComPtr<IOleWindow> pWnd;
			if (m_spInPlaceSite)
				m_spInPlaceSite->QueryInterface(IID_IOleWindow, (void**)&pWnd);
			else if (m_spClientSite)
				m_spClientSite->QueryInterface(IID_IOleWindow, (void**)&pWnd);
			if (pWnd)
			{
				HWND hwnd;
				if (pWnd->GetWindow(&hwnd) == S_OK)
				{
					RECT rect;
					rect.top = rect.bottom = rect.left = rect.right = 0;
					pParam->hWnd = CreateControlWindow(hwnd, rect);
				}
			}
			if (pParam->hWnd == NULL)
				return E_NOINTERFACE;
		}
		pParam->pMgmtInfoBase = m_pMgmtInfoBase;
		HANDLE hSem = pParam->hSem = CreateSemaphore(NULL, 0, 1, NULL);
//		m_sockThread = ::CreateThread(NULL, 0, SockThreadProc, pParam, 0, &m_idThread);
		m_sockThread = (HANDLE)_beginthreadex(NULL, 0, SockThreadProc, pParam, 0, (unsigned int*)&m_idThread);
		if (m_sockThread == NULL)
		{
			delete pParam;
			return E_OUTOFMEMORY;
		}
		WaitForSingleObject(hSem, INFINITE);
		CloseHandle(hSem);
	}
	CSNMPPDU* ppdu = reinterpret_cast<CSNMPPDU*>(p.p);
	LPBYTE pbyte = NULL;
	ULONG len = ppdu->GetPDU(pbyte, this);
	if (len <= 0) return Error(IDS_E_PDU);
	*pReqid = ppdu->GetReqid();
	SendParam* pSendParam = new SendParam(pbyte, len, *pReqid, m_sRetries);
	PostThreadMessage(m_idThread, WM_SOCK_REQ, (WPARAM)pSendParam, 0);

	return S_OK;
}

HRESULT STDMETHODCALLTYPE CSNMPClient::Abort(long reqid)
{
	if (reqid == -1)
		return Close();
	else
		PostThreadMessage(m_idThread, WM_SOCK_ABORT, (WPARAM)reqid, 0);
	return S_OK;
}

LRESULT CSNMPClient::OnError(UINT uMsg, WPARAM wParam, LPARAM lParam, BOOL& bHandled)
{
	Fire_OnError((long)wParam, (long)lParam);
	return TRUE;
}

LRESULT CSNMPClient::OnRecvData(UINT uMsg, WPARAM wParam, LPARAM lParam, BOOL& bHandled)
{
	ISNMPPDU* ppdu;
	HRESULT hr = CoGetInterfaceAndReleaseStream((LPSTREAM)lParam, IID_ISNMPPDU, (void**)&ppdu);
	if (hr == S_OK)
	{
		Fire_OnData((long)wParam, ppdu);
	}
	return TRUE;
}

unsigned int WINAPI CSNMPClient::SockThreadProc(  LPVOID lpParameter  )
{
	CoInitializeEx(NULL, COINIT_APARTMENTTHREADED | COINIT_DISABLE_OLE1DDE);

	ThreadParam* pParam = (ThreadParam*)lpParameter;
	SOCKET socket = pParam->sock;
	sockaddr_in addr = *pParam->addr;
	long lTimeout(pParam->timeout);
	typedef std::list<SendParam*>::iterator SPIter;
	std::list<SendParam*> listSend;
	SPIter SendIt;
	HWND hClient = pParam->hWnd;
	IMgmtInfoBase* pMgmtInfoBase = pParam->pMgmtInfoBase;

	bool fRecvNeed(true);
	WSABUF wsabuff;
	char* pbuffer = (char*)_alloca(PACKET_LENGTH);
	DWORD dwRead, dwFlags, dw;
	WSAOVERLAPPED ov;
	memset(&ov, 0, sizeof(ov));
	ov.hEvent = WSACreateEvent();
	int fromlen;
	sockaddr_in fromaddress;

	MSG msg;
	// create thread message queueu
	::PeekMessage(&msg, NULL, 0, 0, PM_NOREMOVE);
	// activate parent
	ReleaseSemaphore(pParam->hSem, 1, NULL);
	delete pParam;
	//
	for (;;)
	{
		if (fRecvNeed)
		{
			fRecvNeed = false;
			wsabuff.buf = pbuffer;
			wsabuff.len = PACKET_LENGTH;
			fromlen = sizeof(fromaddress);
			dwFlags = dwRead = 0;
			ov.Internal = ov.InternalHigh = ov.Offset = ov.OffsetHigh = 0;
			int n = WSARecvFrom(socket, &wsabuff, 1, &dwRead, &dwFlags,
				(LPSOCKADDR)&fromaddress, &fromlen, &ov, NULL);
			if (n == 0) 
			{
				goto received;
			}
			n = WSAGetLastError();
			if (n != WSA_IO_PENDING)
			{
				::PostMessage(hClient, WM_ERROR, (WPARAM)-1, (LPARAM)n);
				break;
			}
		}
		// check messages are in thread queue
		if (::PeekMessage(&msg, NULL, 0, 0, PM_NOREMOVE))
			dw = WAIT_OBJECT_0 + 1;
		else
			dw = MsgWaitForMultipleObjects(1, &ov.hEvent, FALSE, (listSend.size()) ? 1000 : 600000, QS_ALLINPUT);
		if (dw == WAIT_OBJECT_0 + 1)
		{
			if (GetMessage(&msg, 0, 0, 0) == FALSE)
				break;
			else
			{
				switch (msg.message)
				{
				case WM_SOCK_REQ:
					{
						SendParam* pparam = (SendParam*)msg.wParam;
						ULONG len;
						WSABUF wb;
						len = wb.len = pparam->cbData;
						wb.buf = (char*)pparam->pbData;
						if (WSASendTo(socket, &wb, 1, &len, 0, (LPSOCKADDR)&addr, sizeof(sockaddr_in), NULL, NULL))
						{
							int lasterr = WSAGetLastError();
							::PostMessage(hClient, WM_ERROR, (WPARAM)pparam->reqid, (LPARAM)lasterr);

							delete[] pparam->pbData;
							delete pparam;
						}
						else
						{
							listSend.push_back(pparam);
						}
					}
					break;
				case WM_SOCK_ABORT:
					for (SendIt = listSend.begin(); SendIt != listSend.end(); SendIt++)
					{
						if ((*SendIt)->reqid == (long)msg.wParam)
						{
							delete[] (*SendIt)->pbData;
							delete *SendIt;
							listSend.erase(SendIt);
							break;
						}
					}
					break;
				default:
					break;
				}
			}
		}
		else if (dw == WAIT_TIMEOUT)
		{
			DWORD dwTick = GetTickCount();
			DWORD dwLimit = dwTick - lTimeout;
			for (SendIt = listSend.begin(); SendIt != listSend.end();)
			{
				SendParam* pparam = *SendIt;
				int lasterr(0);
				if (pparam->dwStamp < dwLimit)
				{
					if (pparam->count)
					{
						--pparam->count;
						pparam->dwStamp = dwTick;
						ULONG len;
						WSABUF wb;
						len = wb.len = pparam->cbData;
						wb.buf = (char*)pparam->pbData;
						if (WSASendTo(socket, &wb, 1, &len, 0, (LPSOCKADDR)&addr, sizeof(sockaddr_in), NULL, NULL))
						{
							lasterr = WSAGetLastError();
						}
					}
					else	// retry is over
						lasterr = -1;
					if (lasterr)
					{
						::PostMessage(hClient, WM_ERROR, (WPARAM)pparam->reqid, (LPARAM)lasterr);
						delete[] pparam->pbData;
						delete pparam;
						SendIt = listSend.erase(SendIt);
						continue;
					}
				}
				 SendIt++;
			}
		}
		else
		{
			if (WSAGetOverlappedResult(socket, &ov, &dwRead, TRUE, &dwFlags) == FALSE)
			{
				int n = GetLastError();
				if (n != WSA_IO_PENDING)
				{
					::PostMessage(hClient, WM_ERROR, (WPARAM)-1, (LPARAM)n);
					break;
				}
				fRecvNeed = true;
				continue;
			}
			WSAResetEvent(ov.hEvent);
received:
			if (addr.sin_addr.S_un.S_addr == fromaddress.sin_addr.S_un.S_addr)
			{

				CComObject<CSNMPPDU>* p;
				HRESULT hr = CComObject<CSNMPPDU>::CreateInstance(&p);
				if (hr == S_OK)
				{
					p->put_MgmtInfoBase(pMgmtInfoBase);
					p->SetPDU(wsabuff.len, (LPBYTE)wsabuff.buf);

					bool bFound(false);
					for (SendIt = listSend.begin(); SendIt != listSend.end(); SendIt++)
					{
						if ((*SendIt)->reqid == p->GetReqid())
						{
							delete[] (*SendIt)->pbData;
							delete *SendIt;
							listSend.erase(SendIt);
							bFound = true;
							break;
						}
					}
					if (bFound == false)
					{
						delete p;
					}
					else
					{
						IStream* pstm;
						hr = CoMarshalInterThreadInterfaceInStream(IID_ISNMPPDU, reinterpret_cast<ISNMPPDU*>(p), &pstm);
						if (hr == S_OK)
							::PostMessage(hClient, WM_RECV_DATA, (WPARAM)p->GetReqid(), (LPARAM)pstm);
					}
				}
			}
			fRecvNeed = true;
		}
	}
	WSACloseEvent(ov.hEvent);
	for (SendIt = listSend.begin(); SendIt != listSend.end(); SendIt++)
	{
		delete[] (*SendIt)->pbData;
		delete *SendIt;
	}
	CoUninitialize();
	// It's not necessary to call this.
//	_endthreadex(0);
	// make compiler happy
	return 0;
}

