// SNMPClient.h : Declaration of the CSNMPClient

#ifndef __SNMPCLIENT_H_
#define __SNMPCLIENT_H_

#include "resource.h"       // C V{
#include "SNMPCtlCP.h"
#include "MgmtInfoBase.h"

#define DEFAULT_COMMUNITY   "public"
const int DEFAULT_RETRIES = 3;
const long DEFAULT_TIMEOUT =   30000L;
const short DEFAULT_REMPORT =  SNMP_PORT;
const int DEFAULT_TIME       = 0;
const int DEFAULT_MMS        = 1389; // large, randomly picked for testing purposes

const UINT WM_SOCK_ABORT = WM_APP + 10;
const UINT WM_SOCK_REQ = WM_APP + 11;
const UINT WM_ERROR = WM_APP + 12;
const UINT WM_RECV_DATA = WM_APP + 13;
/////////////////////////////////////////////////////////////////////////////
// CSNMPClient
class ATL_NO_VTABLE CSNMPClient : 
	public CComObjectRootEx<CComSingleThreadModel>,
	public IDispatchImpl<ISNMPClient, &IID_ISNMPClient, &LIBID_SNMPCTLLib>,
	public CComControl<CSNMPClient>,
	public IPersistStreamInitImpl<CSNMPClient>,
	public IOleControlImpl<CSNMPClient>,
	public IOleObjectImpl<CSNMPClient>,
	public IOleInPlaceActiveObjectImpl<CSNMPClient>,
	public IViewObjectExImpl<CSNMPClient>,
	public IOleInPlaceObjectWindowlessImpl<CSNMPClient>,
	public ISupportErrorInfo,
	public IConnectionPointContainerImpl<CSNMPClient>,
	public CComCoClass<CSNMPClient, &CLSID_SNMPClient>,
	public IObjectSafetyImpl<CSNMPClient, INTERFACESAFE_FOR_UNTRUSTED_CALLER | INTERFACESAFE_FOR_UNTRUSTED_DATA>,
	public IProvideClassInfo2Impl<&CLSID_SNMPClient, &DIID__ISNMPClientEvents, &LIBID_SNMPCTLLib>,
	public CProxy_ISNMPClientEvents< CSNMPClient >
{
	friend class CSNMPPDU;
public:
	CSNMPClient()
		: m_sPort(SNMP_PORT), m_sRetries(DEFAULT_RETRIES), m_lTimeout(DEFAULT_TIMEOUT),
		  m_strCommunity("public"), m_sock(INVALID_SOCKET), m_nVersion(SNMP_VERSION_1),
		  m_sockThread(NULL), m_idThread(0), m_pMgmtInfoBase(NULL)
	{
		m_bWindowOnly = TRUE;
		HDC hDC = ::GetDC(NULL);
		int cxPPI = GetDeviceCaps(hDC, LOGPIXELSX);
		int cyPPI = GetDeviceCaps(hDC, LOGPIXELSY);

		m_sizeExtent.cx = (long)MulDiv(2540, GetSystemMetrics(SM_CXICON), cxPPI);
		m_sizeExtent.cy = (long)MulDiv(2540, GetSystemMetrics(SM_CYICON), cyPPI);
		m_sizeNatural = m_sizeExtent;

		::ReleaseDC(NULL, hDC); 
	}

DECLARE_REGISTRY_RESOURCEID(IDR_SNMPCLIENT)

DECLARE_PROTECT_FINAL_CONSTRUCT()

BEGIN_COM_MAP(CSNMPClient)
	COM_INTERFACE_ENTRY(ISNMPClient)
	COM_INTERFACE_ENTRY(IDispatch)
	COM_INTERFACE_ENTRY(IViewObjectEx)
	COM_INTERFACE_ENTRY(IViewObject2)
	COM_INTERFACE_ENTRY(IViewObject)
	COM_INTERFACE_ENTRY(IOleInPlaceObjectWindowless)
	COM_INTERFACE_ENTRY(IOleInPlaceObject)
	COM_INTERFACE_ENTRY2(IOleWindow, IOleInPlaceObjectWindowless)
	COM_INTERFACE_ENTRY(IOleInPlaceActiveObject)
	COM_INTERFACE_ENTRY(IOleControl)
	COM_INTERFACE_ENTRY(IOleObject)
	COM_INTERFACE_ENTRY(IPersistStreamInit)
	COM_INTERFACE_ENTRY2(IPersist, IPersistStreamInit)
	COM_INTERFACE_ENTRY(ISupportErrorInfo)
	COM_INTERFACE_ENTRY(IConnectionPointContainer)
	COM_INTERFACE_ENTRY_IMPL(IConnectionPointContainer)
	COM_INTERFACE_ENTRY(IObjectSafety)
	COM_INTERFACE_ENTRY(IProvideClassInfo)
	COM_INTERFACE_ENTRY(IProvideClassInfo2)
END_COM_MAP()

BEGIN_PROP_MAP(CSNMPClient)
	PROP_DATA_ENTRY("_cx", m_sizeExtent.cx, VT_UI4)
	PROP_DATA_ENTRY("_cy", m_sizeExtent.cy, VT_UI4)
	// Example entries
	// PROP_ENTRY("Property Description", dispid, clsid)
	// PROP_PAGE(CLSID_StockColorPage)
END_PROP_MAP()

BEGIN_CONNECTION_POINT_MAP(CSNMPClient)
	CONNECTION_POINT_ENTRY(DIID__ISNMPClientEvents)
END_CONNECTION_POINT_MAP()

BEGIN_MSG_MAP(CSNMPClient)
	MESSAGE_HANDLER( WM_ERROR, OnError )
	MESSAGE_HANDLER( WM_RECV_DATA, OnRecvData )
	CHAIN_MSG_MAP(CComControl<CSNMPClient>)
	DEFAULT_REFLECTION_HANDLER()
END_MSG_MAP()
// Handler prototypes:
LRESULT OnError(UINT uMsg, WPARAM wParam, LPARAM lParam, BOOL& bHandled);
LRESULT OnRecvData(UINT uMsg, WPARAM wParam, LPARAM lParam, BOOL& bHandled);
//  LRESULT MessageHandler(UINT uMsg, WPARAM wParam, LPARAM lParam, BOOL& bHandled);
//  LRESULT CommandHandler(WORD wNotifyCode, WORD wID, HWND hWndCtl, BOOL& bHandled);
//  LRESULT NotifyHandler(int idCtrl, LPNMHDR pnmh, BOOL& bHandled);

	HRESULT FinalConstruct();

	void FinalRelease();

// ISupportsErrorInfo
	STDMETHOD(InterfaceSupportsErrorInfo)(REFIID riid)
	{
		static const IID* arr[] = 
		{
			&IID_ISNMPClient,
		};
		for (int i=0; i<sizeof(arr)/sizeof(arr[0]); i++)
		{
			if (::InlineIsEqualGUID(*arr[i], riid))
				return S_OK;
		}
		return S_FALSE;
	}

// IViewObjectEx
	DECLARE_VIEW_STATUS(VIEWSTATUS_SOLIDBKGND | VIEWSTATUS_OPAQUE)

// ISNMPClient
	HRESULT STDMETHODCALLTYPE put_PeerName( 
            /* [in] */ BSTR NewVal);
        
	HRESULT STDMETHODCALLTYPE get_PeerName( 
            /* [retval][out] */ BSTR __RPC_FAR *pRetVal);
        
	HRESULT STDMETHODCALLTYPE put_Port( 
            /* [in] */ short NewVal);
        
	HRESULT STDMETHODCALLTYPE get_Port( 
            /* [retval][out] */ short __RPC_FAR *pRetVal);
        
	HRESULT STDMETHODCALLTYPE put_Community( 
            /* [in] */ BSTR NewVal);
        
	HRESULT STDMETHODCALLTYPE get_Community( 
            /* [retval][out] */ BSTR __RPC_FAR *pRetVal);
        
	HRESULT STDMETHODCALLTYPE put_Retries( 
            /* [in] */ short NewVal);
        
	HRESULT STDMETHODCALLTYPE get_Retries( 
            /* [retval][out] */ short __RPC_FAR *pRetVal);
        
	HRESULT STDMETHODCALLTYPE put_Timeout( 
            /* [in] */ long NewVal);
        
	HRESULT STDMETHODCALLTYPE get_Timeout( 
            /* [retval][out] */ long __RPC_FAR *pRetVal);
        
	HRESULT STDMETHODCALLTYPE put_Version( 
            /* [in] */ short NewVal);
        
	HRESULT STDMETHODCALLTYPE get_Version( 
            /* [retval][out] */ short __RPC_FAR *pRetVal);

	HRESULT STDMETHODCALLTYPE put_MgmtInfoBase( 
            /* [in] */ IMgmtInfoBase __RPC_FAR *pNewVal);
        
	HRESULT STDMETHODCALLTYPE get_MgmtInfoBase( 
            /* [retval][out] */ IMgmtInfoBase __RPC_FAR *__RPC_FAR *ppRetVal);

	HRESULT STDMETHODCALLTYPE Open( void);
        
	HRESULT STDMETHODCALLTYPE Close( void);

	HRESULT STDMETHODCALLTYPE NewPDU( 
            /* [retval][out] */ IDispatch __RPC_FAR *__RPC_FAR *ppResult);
        
	HRESULT STDMETHODCALLTYPE Request( 
            /* [in] */ IDispatch __RPC_FAR *pPDU,
            /* [retval][out] */ long __RPC_FAR *pReqid);
        
	HRESULT STDMETHODCALLTYPE Abort(long reqid);
public:

	HRESULT OnDraw(ATL_DRAWINFO& di)
	{
		RECT& rc = *(RECT*)di.prcBounds;
		Rectangle(di.hdcDraw, rc.left, rc.top, rc.right, rc.bottom);

		int cxIcon = GetSystemMetrics(SM_CXICON);
		int cyIcon = GetSystemMetrics(SM_CYICON);
		int x = ((rc.right - rc.left) - cxIcon + 1) / 2;
		int y = ((rc.bottom - rc.top) - cyIcon + 1) / 2;

		HICON hIcon = LoadIcon(_Module.GetModuleInstance(), MAKEINTRESOURCE(IDI_ICON1));
		DrawIcon(di.hdcDraw, rc.left + x, rc.top + y, hIcon);

		return S_OK;
	}

private:
	static int g_Count;	// for winsock library ref count by self.
	static CComMultiThreadModel::AutoCriticalSection s_crit;
	struct ThreadParam {
		SOCKET sock;
		sockaddr_in* addr;
		DWORD timeout;
		HANDLE hSem;
		HWND hWnd;
		IMgmtInfoBase* pMgmtInfoBase;
	};
	struct SendParam {
		SendParam(LPBYTE pb, long cb, long id, int c) : pbData(pb), cbData(cb), reqid(id), count(c), dwStamp(GetTickCount())
		{
		}
		LPBYTE pbData;
		long cbData;
		long reqid;
		int count;
		DWORD dwStamp;
	};
	std::string m_strPeerName;
	inline std::string& GetCommunity() { return m_strCommunity; }
	std::string m_strCommunity;
	short m_sPort;
	short m_sRetries;
	long m_lTimeout;
	SOCKET m_sock;
	sockaddr_in m_addr;
	int m_nVersion;
	HANDLE m_sockThread;
	DWORD m_idThread;
	static unsigned int WINAPI SockThreadProc(  LPVOID lpParameter  );
	IMgmtInfoBase* m_pMgmtInfoBase;
};

#endif //__SNMPCLIENT_H_
