// SNMPPDU.cpp : CSNMPPDU ̃Cve[V
#include "stdafx.h"
#include "SNMPCtl.h"
#include "SNMPClient.h"
#include "SNMPVariable.h"
#include "SNMPPDU.h"
#include "asn1.h"
/////////////////////////////////////////////////////////////////////////////
// CSNMPPDU
//#define SUPPORT_TRAP

ULONG CSNMPPDU::s_uReqid(0);

HRESULT STDMETHODCALLTYPE CSNMPPDU::get_Command( 
            /* [retval][out] */ SNMPCommandConstants __RPC_FAR *pResult)
{
	if (!pResult) return E_POINTER;
	*pResult = 	m_Command;
	return S_OK;
}
        
HRESULT STDMETHODCALLTYPE CSNMPPDU::put_Command( 
            /* [in] */ SNMPCommandConstants NewVal)
{
	if (m_fResponse == false) 
	{
		m_Command = NewVal;
	}
	return S_OK;
}
        
HRESULT STDMETHODCALLTYPE CSNMPPDU::get_ErrorStatus( 
            /* [retval][out] */ SNMPErrorStatusConstants __RPC_FAR *pResult)
{
	if (!pResult) return E_POINTER;
	*pResult = m_Error;
	return S_OK;
}
        
HRESULT STDMETHODCALLTYPE CSNMPPDU::get_ErrorIndex( 
            /* [retval][out] */ long __RPC_FAR *pRetVal)
{
	if (!pRetVal) return E_POINTER;
	*pRetVal = m_ErrorIndex;
	return S_OK;
}
        
HRESULT STDMETHODCALLTYPE CSNMPPDU::put_MgmtInfoBase( 
            /* [in] */ IMgmtInfoBase __RPC_FAR *pNewVal)
{
	if (pNewVal == NULL) return E_POINTER;
	IMgmtInfoBase* pBase;
	HRESULT hr = pNewVal->QueryInterface(IID_IMgmtInfoBase, (void**)&pBase);
	if (hr == S_OK)
	{
		if (m_pMgmtInfoBase)
			m_pMgmtInfoBase->Release();
		m_pMgmtInfoBase = pBase;
	}
	return hr;
}
        
HRESULT STDMETHODCALLTYPE CSNMPPDU::get_MgmtInfoBase( 
            /* [retval][out] */ IMgmtInfoBase __RPC_FAR *__RPC_FAR *ppRetVal)
{
	if (!ppRetVal) return E_POINTER;
	if (m_pMgmtInfoBase == NULL) return Error(IDS_E_NOMIB);
	return m_pMgmtInfoBase->QueryInterface(IID_IMgmtInfoBase, (void**)ppRetVal);
}

HRESULT STDMETHODCALLTYPE CSNMPPDU::get__NewEnum( 
            /* [retval][out] */ IUnknown __RPC_FAR *__RPC_FAR *ppenum)
{
	if (!ppenum) return E_POINTER;
	*ppenum = NULL;
	typedef CComObject<CComEnumOnSTL<IEnumVARIANT, &IID_IEnumVARIANT, VARIANT, _Copy<VARIANT>, std::list<VARIANT> > > enumvar;
	enumvar* p = NULL;
	ATLTRY(p = new enumvar);
	if(p == NULL)
		return E_OUTOFMEMORY;
	HRESULT hRes = p->Init(GetControllingUnknown(), m_listVars);
	if (SUCCEEDED(hRes))
		hRes = p->QueryInterface(IID_IEnumVARIANT, (void**)ppenum);
	if (FAILED(hRes))
		delete p;
	return hRes;
}
        
HRESULT STDMETHODCALLTYPE CSNMPPDU::AddItem( 
            /* [in] */ ISNMPVariable __RPC_FAR *pNewVal)
{
	if (!pNewVal) return E_POINTER;
	if (m_fResponse) return S_OK;

	pNewVal->AddRef();
	VARIANT v;
	VariantInit(&v);
	v.vt = VT_DISPATCH;
	v.pdispVal = pNewVal;
	m_listVars.push_back(v);
	return S_OK;
}
        
HRESULT STDMETHODCALLTYPE CSNMPPDU::InitNew( void)
{
	m_ErrorIndex = 0;
	m_fResponse = false;
	m_uReqid = SNMP_DEFAULT_REQID;
	m_Error = SNMP_ERR_NOERROR;
	m_Command = (SNMPCommandConstants)NULL;
	for (VarsIter it = m_listVars.begin(); it != m_listVars.end(); it++)
		VariantClear(&(*it));
	m_listVars.clear();

	return S_OK;
}

int CSNMPPDU::GetPDU(LPBYTE& pbdst, CSNMPClient* pSession)
{
	BYTE packet[PACKET_LENGTH], apacket[PACKET_LENGTH];
	int length(PACKET_LENGTH);

    if (m_Command == GET_REQ_MSG || m_Command == GETNEXT_REQ_MSG
		|| m_Command == GET_RSP_MSG || m_Command == SET_REQ_MSG 
#ifdef SUPPORT_TRAP
		|| m_Command == TRP_REQ_MSG
#endif
		|| m_Command == GETBULK_REQ_MSG ){
		if (m_uReqid == SNMP_DEFAULT_REQID)
		{
			Lock();
			m_uReqid = ++s_uReqid;
			Unlock();
		}
	}
	else
		return 0;

    LPBYTE cp = packet;
	for (VarsIter vp = m_listVars.begin(); vp != m_listVars.end(); vp++)
	{
		cp = build_var(cp, *vp, length);
		if (cp == NULL) return 0;
	}
	int totallength = cp - packet;

    length = PACKET_LENGTH;
    cp = build_header(apacket, length, (BYTE)(ASN_SEQUENCE | ASN_CONSTRUCTOR), totallength);
    if (cp == NULL)
	{
		return -1;
	}
	CopyMemory(cp, packet, totallength);
    totallength += cp - apacket;

    length = PACKET_LENGTH;
    if (m_Command != TRP_REQ_MSG){
		/* request id */
		cp = build_int(packet, length,
			(BYTE)(ASN_UNIVERSAL | ASN_PRIMITIVE | ASN_INTEGER), (long)m_uReqid);
		if (cp == NULL)
		{
			return -1;
		}
		/* error status */
		cp = build_int(cp, length,
			(BYTE)(ASN_UNIVERSAL | ASN_PRIMITIVE | ASN_INTEGER), 0);
		if (cp == NULL)
		{
			return -1;
		}
		/* error index */
		cp = build_int(cp, length,
			(BYTE)(ASN_UNIVERSAL | ASN_PRIMITIVE | ASN_INTEGER), 0);
		if (cp == NULL)
		{
			return -1;
		}
    } else {	/* this is a trap message */
		/* enterprise */
#ifdef SUPPORT_TRAP
		cp = build_objid(packet, length,
			(BYTE)(ASN_UNIVERSAL | ASN_PRIMITIVE | ASN_OBJECT_ID),
			(oid *)pdu->enterprise, pdu->enterprise_length);
		if (cp == NULL)
		{
			return -1;
		}
		/* agent-addr */
		cp = build_string(cp, length,
			(BYTE)(ASN_UNIVERSAL | ASN_PRIMITIVE | ASN_OCTET_STR),
			(u_char *)&pdu->agent_addr.sin_addr.s_addr, sizeof(pdu->agent_addr.sin_addr.s_addr));
		if (cp == NULL)
		{
			return -1;
		}
		/* generic trap */
		cp = build_int(cp, length,
			(BYTE)(ASN_UNIVERSAL | ASN_PRIMITIVE | ASN_INTEGER),
			(long *)&pdu->trap_type, sizeof(pdu->trap_type));
		if (cp == NULL)
		{
			return -1;
		}
		/* specific trap */
		cp = build_int(cp, length,
			(BYTE)(ASN_UNIVERSAL | ASN_PRIMITIVE | ASN_INTEGER),
			(long *)&pdu->specific_type, sizeof(pdu->specific_type));
		if (cp == NULL)
		{
			return -1;
		}
		/* timestamp  */
		cp = build_int(cp, length,
			(BYTE)(ASN_UNIVERSAL | ASN_PRIMITIVE | ASN_INTEGER),
			(long *)&pdu->time, sizeof(pdu->time));
		if (cp == NULL)
		{
			return -1;
		}
#endif
    }
    if (length < totallength)
		return -1;
	CopyMemory(cp, apacket, totallength);
    totallength += cp - packet;

    length = PACKET_LENGTH;
    cp = build_header(apacket, length, (BYTE)m_Command, totallength);
    if (cp == NULL)
		return -1;
    if (length < totallength)
		return -1;

	CopyMemory(cp, packet, totallength);
    totallength += cp - apacket;

    length = PACKET_LENGTH;

    cp = build_auth( packet, length, totallength, pSession );
    if (cp == NULL)
		return -1;
	length = cp - packet;
    totallength += length;
	pbdst = new BYTE[totallength];
	CopyMemory(pbdst, packet, length);
	CopyMemory(pbdst + length, apacket, totallength - length);

#ifdef SUPP_AUTH
    if( session->qoS & USEC_QOS_AUTH )
	    md5Digest( packet, totallength, cp - (session->contextLen + 16),
		cp - (session->contextLen + 16) );
#endif
	return totallength;
}

void CSNMPPDU::SetPDU(int length, LPBYTE data)
{
	long type;
    long    version;
	std::string strCommunity;
    struct variable_list *vp = 0;
    u_char  *origdata = data;
    int      origlength = length;
    int      ret = 0;

    /* authenticates message and returns length if valid */
    data = parse_auth(data, length, strCommunity, version);
    if (data == NULL)
		return;

	if( version != SNMP_VERSION_1 && version != SNMP_VERSION_2C && version != SNMP_VERSION_2 ) {
	   ATLTRACE(_T("Wrong version: %ld\n"), version);
	   return;
    }

    LPBYTE save_data = data;

	long msg_type(0);
    data = parse_header(data, length, msg_type);
    if (data == NULL)
		return;
    m_Command = (SNMPCommandConstants)msg_type;

#ifdef SUPPORT_AUTH
    if( session->authenticator ) {
		ret = session->authenticator( origdata, origlength, save_data - community_length, 
			community_length, session, pdu );
		if( ret < 0 ) return ret;
    }
#endif

    if (m_Command != TRP_REQ_MSG){
		data = parse_int(data, length, type, msg_type);
		if (data == NULL)
			return;
		m_uReqid = (ULONG)msg_type;
		data = parse_int(data, length, type, msg_type);
		if (data == NULL)
			return;
		m_Error = (SNMPErrorStatusConstants)msg_type;
		data = parse_int(data, length, type, m_ErrorIndex);
		if (data == NULL)
			return;
    } else {
#ifdef SUPPORT_TRAP
		pdu->enterprise_length = MAX_NAME_LEN;
		data = asn_parse_objid(data, &length, &type, objid, &pdu->enterprise_length);
		if (data == NULL)
			return -1;
		pdu->enterprise = (oid *)calloc(1, pdu->enterprise_length * sizeof(oid));
		bcopy((char *)objid, (char *)pdu->enterprise, pdu->enterprise_length * sizeof(oid));

		four = 4;
		data = asn_parse_string(data, &length, &type, (u_char *)&pdu->agent_addr.sin_addr.s_addr, &four);
		if (data == NULL)
			return -1;
		data = asn_parse_int(data, &length, &type, (long *)&pdu->trap_type, sizeof(pdu->trap_type));
		if (data == NULL)
			return -1;
		data = asn_parse_int(data, &length, &type, (long *)&pdu->specific_type, sizeof(pdu->specific_type));
		if (data == NULL)
			return -1;
		data = asn_parse_int(data, &length, &type, (long *)&pdu->time, sizeof(pdu->time));
		if (data == NULL)
			return -1;
#endif
    }
    data = parse_header(data, length, type);
    if (data == NULL)
		return;
    if (type != (ASN_SEQUENCE | ASN_CONSTRUCTOR))
		return;
    while(length > 0){
		CComObject<CSNMPVariable>* p;
		HRESULT hr = CComObject<CSNMPVariable>::CreateInstance(&p);
		if (hr != S_OK)
		{
			ATLTRACE(_T("Create variable object was failed %08X\n"), hr);
			return;
		}
		if (m_pMgmtInfoBase)
			p->put_MgmtInfoBase(m_pMgmtInfoBase);
		VARIANT v;
		VariantInit(&v);
		v.vt = VT_DISPATCH;
		p->QueryInterface(IID_IDispatch, (void**)&v.pdispVal);
		m_listVars.push_back(v);
		data = parse_var(data, p, type, length);
		if (data == NULL)
			return;
    }
}

LPBYTE CSNMPPDU::build_auth(LPBYTE data, int& length, int messagelen, CSNMPClient* session)
{
	LPCSTR params;
	int     plen;

	short version;
	session->get_Version(&version);
	if( version == SNMP_VERSION_2 ) {
#ifdef SUPPORT_V2
		BYTE buff[138]; /* max len of param string */
		LPBYTE pp = buff;
		u_long tmp;

		*pp++ = SNMP_USEC_MODEL;	/* usec model */

		*pp++ = session->qoS;

		memcpy( pp, session->agentID, 12 );
		pp += 12;

		if( is_agent || session->qoS & USEC_QOS_AUTH ) {
			tmp = session->agentBoots;
			*pp++ = (tmp >> 24) & 0xff;
			*pp++ = (tmp >> 16) & 0xff;
			*pp++ = (tmp >> 8) & 0xff;
			*pp++ = tmp & 0xff;
		} else {
			memset( pp, 0, 4 );
			pp += 4;
		}

		if( is_agent || session->qoS & USEC_QOS_AUTH ) {
			tmp = session->agentClock + time(NULL);
			*pp++ = (tmp >> 24) & 0xff;
			*pp++ = (tmp >> 16) & 0xff;
			*pp++ = (tmp >> 8) & 0xff;
			*pp++ = tmp & 0xff;
		} else {
			*(long*)pp = 0;
			pp += 4;
		}

		tmp = session->MMS;
		*pp++ = (tmp >> 8) & 0xff;
		*pp++ = tmp & 0xff;
		
		*pp++ = session->userLen;
		memcpy( pp, session->userName, session->userLen );
		pp += session->userLen;

		if( session->qoS & USEC_QOS_AUTH ) {
			*pp++ = 16;
			memcpy( pp, session->authKey, 16 );
			pp += 16;
		} else {
			*pp++ = 0;
		}

		memcpy( pp, session->contextSelector, session->contextLen );
		pp += session->contextLen;

		params = buff;
		plen = pp - buff;
#endif
	} else {
		std::string& str = session->GetCommunity();
		params = str.c_str();
		plen = str.length();
	}

	data = build_sequence(data, length, (BYTE)(ASN_SEQUENCE | ASN_CONSTRUCTOR), 
	    messagelen + plen + 5);
	if (data == NULL){
		return NULL;
	}
	data = build_int(data, length,
		(BYTE)(ASN_UNIVERSAL | ASN_PRIMITIVE | ASN_INTEGER), version);
	if (data == NULL){
		return NULL;
	}

	data = build_string(data, length,
	    (BYTE)(ASN_UNIVERSAL | ASN_PRIMITIVE | ASN_OCTET_STR), params, plen );
	if (data == NULL){
		return NULL;
	}

	return data;
}

LPBYTE CSNMPPDU::build_var(LPBYTE data, VARIANT& v, int& listlength)
{
	CComQIPtr<ISNMPVariable, &IID_ISNMPVariable> pv(v.pdispVal);
	if (pv == NULL) return NULL;

    int		    dummyLen, headerLen;

    dummyLen = listlength;
	LPBYTE dataPtr = data;
    data += 4;
    dummyLen -=4;
    if (dummyLen < 0)
		return NULL;

    headerLen = data - dataPtr;
    listlength -= headerLen;
	VARIANT vvar;
	VariantInit(&vvar);
	pv->get_ObjId(&vvar);
	data = build_objid(data, listlength, (BYTE)(ASN_UNIVERSAL | ASN_PRIMITIVE | ASN_OBJECT_ID), vvar);
    if (data == NULL) 
		return NULL;

	ASN1Constants vtype;
	pv->get_Type(&vtype);
	VARIANT vVal;
	VariantInit(&vVal);
	pv->get_Value(&vVal);
    switch(vtype){
	case ASN_INTEGER:
	    data = build_int(data, listlength, vtype, vVal);
	    break;
	case ASN_GAUGE:
	case ASN_COUNTER:
	case ASN_TIMETICKS:
	case ASN_UINTEGER:
	    data = build_unsigned_int(data, listlength, vtype, vVal);
	    break;
	case ASN_COUNTER64:
	    data = build_unsigned_int64(data, listlength, vtype, vVal);
	    break;
	case ASN_OCTET_STR:
	case ASN_IPADDRESS:
	case ASN_OPAQUE:
	case ASN_NSAP:
	    data = build_string(data, listlength, vtype, vVal);
	    break;
	case ASN_OBJECT_ID:
	    data = build_objid(data, listlength, vtype, vVal);
	    break;
	case ASN_NULL:
	    data = build_null(data, listlength, vtype);
	    break;
	case ASN_BIT_STR:
//	    data = asn_build_bitstring(data, listlength, var_val_type,
//		    var_val, var_val_len);
		// not supported now.
		data = NULL;
	    break;
	case SNMP_NOSUCHOBJECT:
	case SNMP_NOSUCHINSTANCE:
	case SNMP_ENDOFMIBVIEW:
	    data = build_null(data, listlength, vtype);
	    break;
	default:
	    data = NULL;
		break;
    }
    if (data == NULL){
		return NULL;
    }
    dummyLen = (data - dataPtr) - headerLen;

	build_sequence(dataPtr, dummyLen, (BYTE)(ASN_SEQUENCE | ASN_CONSTRUCTOR), dummyLen);
	return data;
}

LPBYTE CSNMPPDU::build_objid(LPBYTE data, int& datalength, BYTE type, VARIANT& vvar)
{
	int objidlength;
	ULONG* objid;
	if (vvar.vt == (VT_ARRAY | VT_I4))
	{
		objidlength = vvar.parray->rgsabound[0].cElements;
		objid = (ULONG*)vvar.parray->pvData;
	}
	else if (vvar.vt == (VT_ARRAY | VT_VARIANT))
	{
		objidlength = vvar.parray->rgsabound[0].cElements;
		ULONG* oids = (ULONG*)_alloca(objidlength * sizeof(ULONG));
		VARIANT* pv = (VARIANT*)vvar.parray->pvData;

		for (int i = 0; i < objidlength; i++, pv++)
			oids[i] = pv->lVal;
		objid = oids;
	}
	else
		return NULL;
	BYTE buf[MAX_OID_LEN];
	LPBYTE bp = buf;
	ULONG* op = objid;
	ULONG subid, mask, testmask;
	int bits, testbits;

    if (objidlength < 2){
		*bp++ = 0;
		objidlength = 0;
    } else {
		*bp++ = op[1] + (op[0] * 40);
		objidlength -= 2;
		op += 2;
    }

    while(objidlength-- > 0){
		subid = *op++;
		if (subid < 127){ /* off by one? */
			*bp++ = subid;
		} else {
			mask = 0x7F; /* handle subid == 0 case */
			bits = 0;
			/* testmask *MUST* !!!! be of an unsigned type */
			for(testmask = 0x7F, testbits = 0; testmask != 0;testmask <<= 7, testbits += 7){
				if (subid & testmask){	/* if any bits set */
					mask = testmask;
					bits = testbits;
				}
			}
			/* mask can't be zero here */
			for(;mask != 0x7F; mask >>= 7, bits -= 7){
				/* fix a mask that got truncated above */
				if (mask == 0x1E00000)  
					mask = 0xFE00000;
				*bp++ = (BYTE)(((subid & mask) >> bits) | ASN_BIT8);
			}
			*bp++ = (BYTE)(subid & mask);
		}
    }
    int asnlength = bp - buf;
    data = build_header(data, datalength, type, asnlength);
    if (data == NULL)
		return NULL;
    if (datalength < asnlength)
		return NULL;
	CopyMemory(data, buf, asnlength);
    datalength -= asnlength;
    return data + asnlength;
}

/*
 * asn_build_int - builds an ASN object containing an integer.
 *  On entry, datalength is input as the number of valid bytes following
 *   "data".  On exit, it is returned as the number of valid bytes
 *   following the end of this object.
 *
 *  Returns a pointer to the first byte past the end
 *   of this object (i.e. the start of the next object).
 *  Returns NULL on any error.
 */
LPBYTE CSNMPPDU::build_int(LPBYTE data, int& datalength, BYTE type, VARIANT& vVar)
{
/*
 * ASN.1 integer ::= 0x02 asnlength byte {byte}*
 */
	VARIANT vInt;
	VariantInit(&vInt);
	if (VariantChangeType(&vInt, &vVar, 0, VT_I4) != S_OK) return NULL;

	return build_int(data, datalength, type, vInt.lVal);
}

LPBYTE CSNMPPDU::build_int(LPBYTE data, int& datalength, BYTE type, long integer)
{
/*
 * ASN.1 integer ::= 0x02 asnlength byte {byte}*
 */
	int intsize(sizeof(long));
    /*
     * Truncate "unnecessary" bytes off of the most significant end of this
     * 2's complement integer.  There should be no sequence of 9
     * consecutive 1's or 0's at the most significant end of the
     * integer.
     */
    ULONG mask(0x1FF << ((8 * (sizeof(long) - 1)) - 1));

    /* mask is 0xFF800000 on a big-endian machine */
    while((((integer & mask) == 0) || ((integer & mask) == mask))
	  && intsize > 1){
		intsize--;
		integer <<= 8;
    }
    data = build_header(data, datalength, type, intsize);
    if (data == NULL)
		return NULL;
    if (datalength < intsize)
		return NULL;
    datalength -= intsize;
    mask = 0xFF << (8 * (sizeof(long) - 1));
    /* mask is 0xFF000000 on a big-endian machine */
    while(intsize--){
		*data++ = (u_char)((integer & mask) >> (8 * (sizeof(long) - 1)));
		integer <<= 8;
    }
    return data;
}
/*
 * asn_build_unsigned_int - builds an ASN object containing an integer.
 *  On entry, datalength is input as the number of valid bytes following
 *   "data".  On exit, it is returned as the number of valid bytes
 *   following the end of this object.
 *
 *  Returns a pointer to the first byte past the end
 *   of this object (i.e. the start of the next object).
 *  Returns NULL on any error.
 */
LPBYTE CSNMPPDU::build_unsigned_int(LPBYTE data, int& datalength, BYTE type, VARIANT& vVar)
{
	VARIANT vInt;
	VariantInit(&vInt);
	if (VariantChangeType(&vInt, &vVar, 0, VT_I4) != S_OK) return NULL;

    ULONG integer = vInt.lVal;
	int intsize(sizeof(ULONG));

    bool add_null_byte(false);

    ULONG mask(0xFF << (8 * (sizeof(long) - 1)));
    /* mask is 0xFF000000 on a big-endian machine */
    if ((u_char)((integer & mask) >> (8 * (sizeof(long) - 1))) & 0x80){
		/* if MSB is set */
		add_null_byte = true;
		intsize++;
    }
    /*
     * Truncate "unnecessary" bytes off of the most significant end of this 2's complement integer.
     * There should be no sequence of 9 consecutive 1's or 0's at the most significant end of the
     * integer.
     */
    mask = 0x1FF << ((8 * (sizeof(long) - 1)) - 1);
    /* mask is 0xFF800000 on a big-endian machine */
    while((((integer & mask) == 0) || ((integer & mask) == mask)) && intsize > 1){
		intsize--;
		integer <<= 8;
    }
    data = build_header(data, datalength, type, intsize);
    if (data == NULL)
		return NULL;
    if (datalength < intsize)
		return NULL;
    datalength -= intsize;
    if (add_null_byte){
		*data++ = '\0';
		intsize--;
    }
    mask = 0xFF << (8 * (sizeof(long) - 1));
    /* mask is 0xFF000000 on a big-endian machine */
    while(intsize--){
		*data++ = (u_char)((integer & mask) >> (8 * (sizeof(long) - 1)));
		integer <<= 8;
    }
    return data;
}

/*
 * asn_build_string - Builds an ASN octet string object containing the input string.
 *  On entry, datalength is input as the number of valid bytes following
 *   "data".  On exit, it is returned as the number of valid bytes
 *   following the beginning of the next object.
 *
 *  Returns a pointer to the first byte past the end
 *   of this object (i.e. the start of the next object).
 *  Returns NULL on any error.
 */
LPBYTE CSNMPPDU::build_string(LPBYTE data, int& datalength, BYTE type, VARIANT& vVar)
{
	VARIANT vStr;
	VariantInit(&vStr);
	if (VariantChangeType(&vStr, &vVar, 0, VT_BSTR) != S_OK) return NULL;
	USES_CONVERSION;
	LPSTR ps = W2A(vStr.bstrVal);
	VariantClear(&vStr);
	return build_string(data, datalength, type, ps, lstrlenA(ps));
}

LPBYTE CSNMPPDU::build_string(LPBYTE data, int& datalength, BYTE type, LPCSTR str, int strlength)
{
	LPBYTE pb = (LPBYTE)str;
/*
 * ASN.1 octet string ::= primstring | cmpdstring
 * primstring ::= 0x04 asnlength byte {byte}*
 * cmpdstring ::= 0x24 asnlength string {string}*
 * This code will never send a compound string.
 */
    data = build_header(data, datalength, type, strlength);
    if (data == NULL)
		return NULL;
    if (datalength < strlength)
		return NULL;
	CopyMemory(data, pb, strlength);
    datalength -= strlength;
    return data + strlength;
}
/*
 * asn_build_unsigned_int64 - builds an ASN object containing a 64 bit integer.
 *  On entry, datalength is input as the number of valid bytes following
 *   "data".  On exit, it is returned as the number of valid bytes
 *   following the end of this object.
 *
 *  Returns a pointer to the first byte past the end
 *   of this object (i.e. the start of the next object).
 *  Returns NULL on any error.
 */
LPBYTE CSNMPPDU::build_unsigned_int64(LPBYTE data, int& datalength, BYTE type, VARIANT& vVar)
{
	VARIANT vInt;
	VariantInit(&vInt);
	if (VariantChangeType(&vInt, &vVar, 0, VT_CY) != S_OK) return NULL;
	
	unsigned __int64 integer = *(unsigned __int64*)&vInt.cyVal;
	integer /= 10000;
	int intsize(sizeof(__int64));

    bool add_null_byte(false);

    unsigned __int64 mask(0xFF << (8 * (sizeof(__int64) - 1)));
    /* mask is 0xFF000000 on a big-endian machine */
    if ((u_char)((integer & mask) >> (8 * (sizeof(__int64) - 1))) & 0x80){
		/* if MSB is set */
		add_null_byte = true;
		intsize++;
    }
    /*
     * Truncate "unnecessary" bytes off of the most significant end of this 2's complement integer.
     * There should be no sequence of 9 consecutive 1's or 0's at the most significant end of the
     * integer.
     */
    mask = 0x1FF << ((8 * (sizeof(__int64) - 1)) - 1);
    /* mask is 0xFF800000 on a big-endian machine */
    while((((integer & mask) == 0) || ((integer & mask) == mask)) && intsize > 1){
		intsize--;
		integer <<= 8;
    }
    data = build_header(data, datalength, type, intsize);
    if (data == NULL)
		return NULL;
    if (datalength < intsize)
		return NULL;
    datalength -= intsize;
    if (add_null_byte){
		*data++ = '\0';
		intsize--;
    }
    mask = 0xFF << (8 * (sizeof(__int64) - 1));
    /* mask is 0xFF000000 on a big-endian machine */
    while(intsize--){
		*data++ = (u_char)((integer & mask) >> (8 * (sizeof(__int64) - 1)));
		integer <<= 8;
    }
    return data;
}

/*
 * asn_build_header - builds an ASN header for an object with the ID and
 * length specified.
 *  On entry, datalength is input as the number of valid bytes following
 *   "data".  On exit, it is returned as the number of valid bytes
 *   in this object following the id and length.
 *
 *  This only works on data types < 30, i.e. no extension octets.
 *  The maximum length is 0xFFFF;
 *
 *  Returns a pointer to the first byte of the contents of this object.
 *  Returns NULL on any error.
 */
LPBYTE CSNMPPDU::build_header(LPBYTE data, int& datalength, BYTE type, int length)
{
    if (datalength < 1)
		return NULL;
    *data++ = type;
    datalength--;
    return build_length(data, datalength, length);
}

LPBYTE CSNMPPDU::build_length(LPBYTE data, int& datalength, int length)
{
	LPBYTE start_data = data;

    /* no indefinite lengths sent */
    if (length < 0x80){
		if (datalength < 1){
			return NULL;
		}	    
		*data++ = (BYTE)length;
    } else if (length <= 0xFF){
		if (datalength < 2){
			return NULL;
		}	    
		*data++ = (BYTE)(0x01 | ASN_LONG_LEN);
		*data++ = (BYTE)length;
    } else { /* 0xFF < length <= 0xFFFF */
		if (datalength < 3){
			return NULL;
		}	    
		*data++ = (BYTE)(0x02 | ASN_LONG_LEN);
		*data++ = (BYTE)((length >> 8) & 0xFF);
		*data++ = (BYTE)(length & 0xFF);
    }
    datalength -= (data - start_data);
    return data;
}

/*
 * asn_build_sequence - builds an ASN header for a sequence with the ID and
 * length specified.
 *  On entry, datalength is input as the number of valid bytes following
 *   "data".  On exit, it is returned as the number of valid bytes
 *   in this object following the id and length.
 *
 *  This only works on data types < 30, i.e. no extension octets.
 *  The maximum length is 0xFFFF;
 *
 *  Returns a pointer to the first byte of the contents of this object.
 *  Returns NULL on any error.
 */
LPBYTE CSNMPPDU::build_sequence(LPBYTE data, int& datalength, BYTE type, int length)
{
    datalength -= 4;
    if (datalength < 0){
		datalength += 4;	/* fix up before punting */
		return NULL;
    }
    *data++ = type;
    *data++ = (BYTE)(0x02 | ASN_LONG_LEN);
    *data++ = (BYTE)((length >> 8) & 0xFF);
	*data++ = (BYTE)(length & 0xFF);
    return data;
}

/*
 * asn_build_null - Builds an ASN null object.
 *  On entry, datalength is input as the number of valid bytes following
 *   "data".  On exit, it is returned as the number of valid bytes
 *   following the beginning of the next object.
 *
 *  Returns a pointer to the first byte past the end
 *   of this object (i.e. the start of the next object).
 *  Returns NULL on any error.
 */
LPBYTE CSNMPPDU::build_null(LPBYTE data, int& datalength, BYTE type)
{
/*
 * ASN.1 null ::= 0x05 0x00
 */
    return build_header(data, datalength, type, 0);
}

LPBYTE CSNMPPDU::parse_auth(LPBYTE data, int& length, std::string& sid, long& version)
{
	long type;
	data = parse_header(data, length, type);
    if (data == NULL){
		return NULL;
    }
    if (type != (ASN_SEQUENCE | ASN_CONSTRUCTOR)){
		return NULL;
    }
    data = parse_int(data, length, type, version);
    if (data == NULL){
		return NULL;
    }
    data = parse_string(data, length, type, sid);
    if (data == NULL){
		return NULL;
    }

    return data;
}

/*
 * asn_parse_header - interprets the ID and length of the current object.
 *  On entry, datalength is input as the number of valid bytes following
 *   "data".  On exit, it is returned as the number of valid bytes
 *   in this object following the id and length.
 *
 *  Returns a pointer to the first byte of the contents of this object.
 *  Returns NULL on any error.
 */
LPBYTE CSNMPPDU::parse_header(LPBYTE data, int& datalength, long& type)
{
	LPBYTE bufp = data;
	int asn_length(0);

    /* this only works on data types < 30, i.e. no extension octets */
    if (IS_EXTENSION_ID(*bufp)){
		return NULL;
    }
	type = *bufp;
    bufp = parse_length(bufp + 1, asn_length);
    if (bufp == NULL)
		return NULL;
    int header_len = bufp - data;
    if (header_len + asn_length > datalength){
		return NULL;
    }
    datalength = asn_length;
    return bufp;
}

/*
 * asn_parse_length - interprets the length of the current object.
 *  On exit, length contains the value of this length field.
 *
 *  Returns a pointer to the first byte after this length
 *  field (aka: the start of the data field).
 *  Returns NULL on any error.
 */
LPBYTE CSNMPPDU::parse_length(LPBYTE data, int& length)
{
	BYTE lengthbyte = *data;

    if (lengthbyte & ASN_LONG_LEN){
		lengthbyte &= ~ASN_LONG_LEN;	/* turn MSb off */
		if (lengthbyte == 0){
			return NULL;
		}
		if (lengthbyte > sizeof(long)){
			return NULL;
		}
		CopyMemory(&length, data + 1, lengthbyte);
		length = ntohl(length);
		length >>= (8 * ((sizeof length) - lengthbyte));
		return data + lengthbyte + 1;
    } else { /* short asnlength */
		length = lengthbyte;
		return data + 1;
    }
}

/*
 * asn_parse_int - pulls a long out of an ASN int type.
 *  On entry, datalength is input as the number of valid bytes following
 *   "data".  On exit, it is returned as the number of valid bytes
 *   following the end of this object.
 *
 *  Returns a pointer to the first byte past the end
 *   of this object (i.e. the start of the next object).
 *  Returns NULL on any error.
 */
LPBYTE CSNMPPDU::parse_int(LPBYTE data, int& datalength, long& type, long& intp)
{
/*
 * ASN.1 integer ::= 0x02 asnlength byte {byte}*
 */
	LPBYTE bufp = data;
	int asn_length(0);
	intp = 0;

    type = *bufp++;
    bufp = parse_length(bufp, asn_length);
    if (bufp == NULL){
		return NULL;
    }
    if (asn_length + (bufp - data) > datalength){
		return NULL;
    }
    if (asn_length > sizeof(long)){
		return NULL;
    }
    datalength -= asn_length + (bufp - data);
    if (*bufp & 0x80)
		intp = -1; /* integer is negative */
    while(asn_length--)
		intp = (intp << 8) | *bufp++;
    return bufp;
}

/*
 * asn_parse_string - pulls an octet string out of an ASN octet string type.
 *  On entry, datalength is input as the number of valid bytes following
 *   "data".  On exit, it is returned as the number of valid bytes
 *   following the beginning of the next object.
 *
 *  "string" is filled with the octet string.
 *
 *  Returns a pointer to the first byte past the end
 *   of this object (i.e. the start of the next object).
 *  Returns NULL on any error.
 */
LPBYTE CSNMPPDU::parse_string(LPBYTE data, int& datalength, long& type, std::string& string)
{
/*
 * ASN.1 octet string ::= primstring | cmpdstring
 * primstring ::= 0x04 asnlength byte {byte}*
 * cmpdstring ::= 0x24 asnlength string {string}*
 */
	LPBYTE bufp = data;
	int asn_length(0);

    type = *bufp++;
    bufp = parse_length(bufp, asn_length);
    if (bufp == NULL)
		return NULL;
    if (asn_length + (bufp - data) > datalength){
		return NULL;
    }
	string.append((char*)bufp, asn_length);
	datalength -= (int)asn_length + (bufp - data);
    return bufp + asn_length;
}

LPBYTE CSNMPPDU::parse_var(LPBYTE data, ISNMPVariable* pvar, long& type, int& listlength)
{
    u_char	    *var_op_start = data;

	long var_op_type(0);
	int var_op_len(listlength);
    data = parse_header(data, var_op_len, var_op_type);
    if (data == NULL){
		ATLTRACE(_T("snmp_parse_var_op(): 1 asn_parse_header() == NULL\n"));
		return NULL;
    }
    if (var_op_type != (ASN_SEQUENCE | ASN_CONSTRUCTOR))
		return NULL;
	VARIANT var;
	VariantInit(&var);
    data = parse_objid(data, var_op_len, var_op_type, var);

    if (data == NULL){
		ATLTRACE(_T("snmp_parse_var_op(): asn_parse_objid() == NULL\n"));
		return NULL;
    }
    if (var_op_type != (ASN_UNIVERSAL | ASN_PRIMITIVE | ASN_OBJECT_ID))
		return NULL;
	pvar->put_ObjId(var);
    /* find out what type of object this is */
	LPBYTE pvdata = data;	// save
	int len(var_op_len);
    data = parse_header(data, var_op_len, type);
    if (data == NULL){
		ATLTRACE(_T("snmp_parse_var_op(): 2 asn_parse_header() == NULL\n"));
		return NULL;
    }

	pvar->put_Type((ASN1Constants)type);
	USES_CONVERSION;
	VARIANT v;
	VariantInit(&v);
	switch(type){
	case ASN_INTEGER:
		v.vt = VT_I4;
		parse_int(pvdata, len, type, v.lVal);
		break;
	case ASN_COUNTER:
	case ASN_GAUGE:
	case ASN_TIMETICKS:
	case ASN_UINTEGER:
		{
			ULONG ul;
			parse_unsigned_int(pvdata, len, type, ul);
			v.lVal = (long)ul;
			v.vt = VT_I4;
		}
		
		break;
	case ASN_COUNTER64:
		parse_unsigned_int64(pvdata, len, type, v);
		break;
	case ASN_IPADDRESS:
		{
			std::string str;
			if (parse_string(pvdata, len, type, str))
			{
				TCHAR sz[16];
				wsprintf(sz, _T("%d.%d.%d.%d"), (BYTE)str[0], (BYTE)str[1], (BYTE)str[2], (BYTE)str[3]);
				v.bstrVal = SysAllocString(T2OLE(sz));
				v.vt = VT_BSTR;
			}
		}
		break;
	case ASN_OCTET_STR:
	case ASN_OPAQUE:
	case ASN_NSAP:
		{
			std::string str;
			if (parse_string(pvdata, len, type, str))
			{
#ifdef FORCE_PRINTABLE
				bool fPrintable(true);
				for (std::string::iterator it = str.begin(); it != str.end(); it++)
				{
					if (isgraph(*it) == false)
					{
						fPrintable = false;
						break;
					}	
				}
				if (fPrintable)
				{
					v.bstrVal = SysAllocString(A2W(str.c_str()));
				}
				else
				{
					LPTSTR p = reinterpret_cast<LPTSTR>(_alloca(str.length() * sizeof(TCHAR) * 2 + sizeof(TCHAR)));
					LPTSTR pt = p;
					for (std::string::iterator it = str.begin(); it != str.end(); it++)
						pt += wsprintf(pt, _T("%02X"), (BYTE)*it);
					*pt = _T('\0');
					v.bstrVal = SysAllocString(T2OLE(p));
				}
#else
				LPOLESTR p = reinterpret_cast<LPOLESTR>(_alloca(str.length() * sizeof(OLECHAR)));
				LPOLESTR pt = p;
				for (std::string::iterator it = str.begin(); it != str.end(); it++)
				{
					*pt++ = (BYTE)*it;
				}
				v.bstrVal = SysAllocStringLen(p, str.length());
#endif
				v.vt = VT_BSTR;
			}
		}
		break;
	case ASN_OBJECT_ID:
		parse_objid(pvdata, len, type, v);
		break;
	case SNMP_NOSUCHOBJECT:
	case SNMP_NOSUCHINSTANCE:
	case SNMP_ENDOFMIBVIEW:
	case ASN_NULL:
		break;
	default:
		ATLTRACE(_T("bad type returned (%x)\n"), type);
		break;
	}
	pvar->put_Value(v);
	VariantClear(&v);

    data += var_op_len;
    listlength -= (int)(data - var_op_start);
    return data;
}

/*
 * asn_parse_objid - pulls an object indentifier out of an ASN object identifier type.
 *  On entry, datalength is input as the number of valid bytes following
 *   "data".  On exit, it is returned as the number of valid bytes
 *   following the beginning of the next object.
 *
 *  "objid" is filled with the object identifier.
 *
 *  Returns a pointer to the first byte past the end
 *   of this object (i.e. the start of the next object).
 *  Returns NULL on any error.
 */
LPBYTE CSNMPPDU::parse_objid(LPBYTE data, int& datalength, long& type, VARIANT& var)
{
/*
 * ASN.1 objid ::= 0x06 asnlength subidentifier {subidentifier}*
 * subidentifier ::= {leadingbyte}* lastbyte
 * leadingbyte ::= 1 7bitvalue
 * lastbyte ::= 0 7bitvalue
 */
	LPBYTE bufp = data;
	oid objid[MAX_NAME_LEN];
	oid* oidp = &objid[1];
	int objidlength(MAX_NAME_LEN);
	int asn_length(0);

    type = *bufp++;
    bufp = parse_length(bufp, asn_length);
    if (bufp == NULL)
		return NULL;
    if (asn_length + (bufp - data) > datalength){
		return NULL;
    }
	datalength -= (int)asn_length + (bufp - data);

    /* Handle invalid object identifier encodings of the form 06 00 robustly */
    if (asn_length == 0)
		objid[0] = objid[1] = 0;

	ULONG subidentifier;
	int length(asn_length);
    objidlength--;	/* account for expansion of first byte */
    while (length > 0 && objidlength-- > 0){
		subidentifier = 0;
		do {	/* shift and add in low order 7 bits */
			subidentifier = (subidentifier << 7) + (*bufp & ~ASN_BIT8);
			length--;
		} while (*bufp++ & ASN_BIT8);	/* last byte has high bit clear */
		if (subidentifier > (u_long)MAX_SUBID){
			ATLTRACE(_T("subidentifier too long\n"));
			return NULL;
		}
		*oidp++ = (oid)subidentifier;
    }

    /*
     * The first two subidentifiers are encoded into the first component
     * with the value (X * 40) + Y, where:
     *	X is the value of the first subidentifier.
     *  Y is the value of the second subidentifier.
     */
    subidentifier = (u_long)objid[1];
    if (subidentifier == 0x2B){
		objid[0] = 1;
		objid[1] = 3;
    } else {
		objid[1] = (u_char)(subidentifier % 40);
		objid[0] = (u_char)((subidentifier - objid[1]) / 40);
    }

    SAFEARRAYBOUND rgsabound[1];    
	rgsabound[0].lLbound = 0;
    rgsabound[0].cElements = (int)(oidp - objid);
	var.parray = SafeArrayCreate(VT_VARIANT, 1, rgsabound);
	if (var.parray == NULL) return NULL;
	var.vt = VT_ARRAY | VT_VARIANT;
	VARIANT* vp = (VARIANT*)var.parray->pvData;
	for (int i = 0; i < rgsabound[0].cElements; i++, vp++)
	{
		VariantInit(vp);
		vp->vt = VT_I4;
		vp->lVal = objid[i];
	}
    return bufp;
}

/*
 * asn_parse_unsigned_int - pulls an unsigned long out of an ASN int type.
 *  On entry, datalength is input as the number of valid bytes following
 *   "data".  On exit, it is returned as the number of valid bytes
 *   following the end of this object.
 *
 *  Returns a pointer to the first byte past the end
 *   of this object (i.e. the start of the next object).
 *  Returns NULL on any error.
 */
LPBYTE CSNMPPDU::parse_unsigned_int(LPBYTE data, int& datalength, long& type, ULONG& intp)
{
	long lVal(0);
	LPBYTE pb = parse_int(data, datalength, type, lVal);
	if (pb)
	{
		intp = (ULONG)lVal;
	}
	return pb;
}

/*
 * asn_parse_unsigned_int64 - pulls a 64 bit unsigned long out of an ASN int
 * type.
 *  On entry, datalength is input as the number of valid bytes following
 *   "data".  On exit, it is returned as the number of valid bytes
 *   following the end of this object.
 *
 *  Returns a pointer to the first byte past the end
 *   of this object (i.e. the start of the next object).
 *  Returns NULL on any error.
 */
LPBYTE CSNMPPDU::parse_unsigned_int64(LPBYTE data, int& datalength, long& type, VARIANT& v)
{
/*
 * ASN.1 integer ::= 0x02 asnlength byte {byte}*
 */
	LPBYTE bufp = data;
	int asn_length(0);
    int intsize = sizeof(__int64);
	__int64 i64(0);
 
	type = *bufp++;
    bufp = parse_length(bufp, asn_length);
    if (bufp == NULL){
		return NULL;
    }
    if (asn_length + (bufp - data) > datalength){
		return NULL;
    }
    if (asn_length > sizeof(__int64)) {
		return NULL;
    }
    datalength -= asn_length + (bufp - data);
    if (*bufp & 0x80){
		i64 = -1;
    }
    while(asn_length--){
		i64 = (i64 << 8) | *bufp++;
    }
	i64 *= 10000;
	v.vt = VT_CY;
	*(__int64*)&v.cyVal = i64;
    return bufp;
}

