// mib.c
// 
#include <stdafx.h>
#include "mib.h"

#pragma warning(disable : 4005)
////////////////////////////////////////////////////////////////////////////////////////
/* non-aggregate types for tree end nodes */
#define TYPE_OTHER	    0
#define TYPE_OBJID	    1
#define TYPE_OCTETSTR	    2
#define TYPE_INTEGER	    3
#define TYPE_NETADDR	    4
#define	TYPE_IPADDR	    5
#define TYPE_COUNTER	    6
#define TYPE_GAUGE	    7
#define TYPE_TIMETICKS	    8
#define TYPE_OPAQUE	    9
#define TYPE_NULL	    10
#define TYPE_COUNTER64      11
#define TYPE_BITSTRING      12
#define TYPE_NSAPADDRESS    13
#define TYPE_UINTEGER	    14
////////////////////////////////////////////////////////////////////////////////////////
const int SYNTAX_MASK = 0x80;
/* types of tokens
 Tokens wiht the SYNTAX_MASK bit set are syntax tokens */
#define	CONTINUE    -1
#define ENDOFFILE   0
#define LABEL	    1
#define SUBTREE	    2
#define SYNTAX	    3
#define OBJID	    (4 | SYNTAX_MASK)
#define OCTETSTR    (5 | SYNTAX_MASK)
#define INTEGER	    (6 | SYNTAX_MASK)
#define NETADDR	    (7 | SYNTAX_MASK)
#define	IPADDR	    (8 | SYNTAX_MASK)
#define COUNTER	    (9 | SYNTAX_MASK)
#define GAUGE	    (10 | SYNTAX_MASK)
#define TIMETICKS   (11 | SYNTAX_MASK)
#define OPAQUE	    (12 | SYNTAX_MASK)
#define NUL	    (13 | SYNTAX_MASK)
#define SEQUENCE    14
#define OF	    15	/* SEQUENCE OF */
#define OBJTYPE	    16
#define ACCESS	    17
#define READONLY    18
#define READWRITE   19
#define	WRITEONLY   20
#define NOACCESS    21
#define STATUS	    22
#define MANDATORY   23
#define OPTIONAL    24
#define OBSOLETE    25
/* #define RECOMMENDED 26 */
#define PUNCT	    27
#define EQUALS	    28
#define NUMBER	    29
#define LEFTBRACKET 30
#define RIGHTBRACKET 31
#define	LEFTPAREN   32
#define RIGHTPAREN  33
#define COMMA	    34
#define DESCRIPTION 35
#define QUOTESTRING 36
#define INDEX       37
#define DEFVAL      38
#define DEPRECATED  39
#define SIZE        40
#define BITSTRING   (41 | SYNTAX_MASK)
#define NSAPADDRESS (42 | SYNTAX_MASK)
#define COUNTER64   (43 | SYNTAX_MASK)
#define OBJGROUP    44
#define NOTIFTYPE   45
#define AUGMENTS    46
#define COMPLIANCE  47
#define READCREATE  48
#define UNITS       49
#define REFERENCE   50
#define NUM_ENTRIES 51
#define MODULEIDENTITY 52
#define LASTUPDATED 53
#define ORGANIZATION 54
#define CONTACTINFO 55
#define UINTEGER32 (56 | SYNTAX_MASK)
#define CURRENT	    57
#define DEFINITIONS 58
#define END         59
#define SEMI        60

#ifdef _DEBUG
void CMIB::print_subtree(int count)
{
	CMIB* tp;
    int i;

    for(i = 0; i < count; i++)
		ATLTRACE(_T("  "));
	ATLTRACE(_T("Children of %s:\n"), label.c_str());
    count++;
    for(tp = child_list; tp; tp = tp->next_peer){
		for(i = 0; i < count; i++)
			ATLTRACE(_T("  "));
		ATLTRACE(_T("%s\n"), tp->label.c_str());
    }
    for(tp = child_list; tp; tp = tp->next_peer){
		tp->print_subtree(count);
    }
}
#endif /* TEST */

struct tc {
	tc() : type(0) {}
	enum {
		MAXTC = 128,
	};
	static int get_tc(std::string&, EnumList&);
	static tc* search_tc(int t);
	inline void setDescriptor(std::string& s) { descriptor = s; }
	inline void setType(int t) { type = t; }
	inline void addEnum(EnumInts e) { enums.push_back(e); }
private:
	int type;
	std::string descriptor;
	EnumList enums;
	static tc tclist[];
};

tc tc::tclist[tc::MAXTC];

tc* tc::search_tc(int t)
{
	for (int i = 0; i < MAXTC; i++)
	{
		if (tclist[i].type == t)
			return &tclist[i];
	}
	return NULL;
}

int tc::get_tc(std::string& descriptor, EnumList& ep)
{
    for(int i = 0; i < MAXTC; i++){
		if (tclist[i].type == 0)
			break;
		if (descriptor == tclist[i].descriptor){
			ep = tclist[i].enums;
			return tclist[i].type;
		}
    }
    return LABEL;
}

struct tok {
	static void init();
	enum {
		HASHSIZE = 32,
	};

	inline static int BUCKET(int x) { return x & 0x01f; }
	static int get_token(LPSTR& fp, std::string& token);
	static int parseQuoteString(LPSTR& fp, std::string& token);
	static int tossObjectIdentifier(LPSTR& fp);
	//
	char *name;			/* token name */
	int len;			/* length not counting nul */
	int token;			/* value */
	int hash;			/* hash of name */
	tok *next;		/* pointer to next in hash table */

	static tok *buckets[];
	static tok tokens[];
	static int Line;
};

int tok::Line(1);
tok* tok::buckets[tok::HASHSIZE];
tok tok::tokens[] = {
	{ "obsolete", sizeof ("obsolete")-1, OBSOLETE },
	{ "Opaque", sizeof ("Opaque")-1, OPAQUE },
/*	{ "recommended", sizeof("recommended")-1, RECOMMENDED },  */
	{ "optional", sizeof ("optional")-1, OPTIONAL },
	{ "LAST-UPDATED", sizeof ("LAST-UPDATED")-1, LASTUPDATED },
	{ "ORGANIZATION", sizeof ("ORGANIZATION")-1, ORGANIZATION },
	{ "CONTACT-INFO", sizeof ("CONTACT-INFO")-1, CONTACTINFO },
	{ "MODULE-IDENTITY", sizeof ("MODULE-IDENTITY")-1, MODULEIDENTITY },
	{ "MODULE-COMPLIANCE", sizeof ("MODULE-COMPLIANCE")-1, COMPLIANCE },
    { "DEFINITIONS", sizeof("DEFINITIONS")-1, DEFINITIONS},
	{ "END", sizeof("END")-1, END},
	{ ";", sizeof(";")-1, SEMI},
	{ "AUGMENTS", sizeof ("AUGMENTS")-1, AUGMENTS },
	{ "not-accessible", sizeof ("not-accessible")-1, NOACCESS },
	{ "write-only", sizeof ("write-only")-1, WRITEONLY },
	{ "NsapAddress", sizeof("NsapAddress")-1, NSAPADDRESS},
	{ "UNITS", sizeof("Units")-1, UNITS},
	{ "REFERENCE", sizeof("REFERENCE")-1, REFERENCE},
	{ "NUM-ENTRIES", sizeof("NUM-ENTRIES")-1, NUM_ENTRIES},
	{ "BITSTRING", sizeof("BitString")-1, BITSTRING},
	{ "BIT", sizeof("BIT")-1, CONTINUE},
	{ "Counter64", sizeof("Counter64")-1, COUNTER64},
	{ "TimeTicks", sizeof ("TimeTicks")-1, TIMETICKS },
	{ "NOTIFICATION-TYPE", sizeof ("NOTIFICATION-TYPE")-1, NOTIFTYPE },
	{ "OBJECT-GROUP", sizeof ("OBJECT-GROUP")-1, OBJGROUP },
	{ "OBJECTIDENTIFIER", sizeof ("OBJECTIDENTIFIER")-1, OBJID },
	/*
	 * This CONTINUE appends the next word onto OBJECT,
	 * hopefully matching OBJECTIDENTIFIER above.
	 */
	{ "OBJECT", sizeof ("OBJECT")-1, CONTINUE },
	{ "NetworkAddress", sizeof ("NetworkAddress")-1, NETADDR },
	{ "Gauge", sizeof ("Gauge")-1, GAUGE },
	{ "read-write", sizeof ("read-write")-1, READWRITE },
	{ "read-create", sizeof ("read-create")-1, READCREATE },
	{ "OCTETSTRING", sizeof ("OCTETSTRING")-1, OCTETSTR },
	{ "OCTET", sizeof ("OCTET")-1, -1 },
	{ "OF", sizeof ("OF")-1, OF },
	{ "SEQUENCE", sizeof ("SEQUENCE")-1, SEQUENCE },
	{ "NULL", sizeof ("NULL")-1, NUL },
	{ "IpAddress", sizeof ("IpAddress")-1, IPADDR },
	{ "UInteger32", sizeof ("UInteger32")-1, UINTEGER32 },
	{ "INTEGER", sizeof ("INTEGER")-1, INTEGER },
	{ "Counter", sizeof ("Counter")-1, COUNTER },
	{ "read-only", sizeof ("read-only")-1, READONLY },
	{ "DESCRIPTION", sizeof ("DESCRIPTION")-1, DESCRIPTION },
	{ "INDEX", sizeof ("INDEX")-1, INDEX },
	{ "DEFVAL", sizeof ("DEFVAL")-1, DEFVAL },
	{ "deprecated", sizeof ("deprecated")-1, DEPRECATED },
	{ "SIZE", sizeof ("SIZE")-1, SIZE },
	{ "MAX-ACCESS", sizeof ("MAX-ACCESS")-1, ACCESS },
	{ "ACCESS", sizeof ("ACCESS")-1, ACCESS },
	{ "mandatory", sizeof ("mandatory")-1, MANDATORY },
	{ "current", sizeof ("current")-1, CURRENT },
	{ "STATUS", sizeof ("STATUS")-1, STATUS },
	{ "SYNTAX", sizeof ("SYNTAX")-1, SYNTAX },
	{ "OBJECT-TYPE", sizeof ("OBJECT-TYPE")-1, OBJTYPE },
	{ "{", sizeof ("{")-1, LEFTBRACKET },
	{ "}", sizeof ("}")-1, RIGHTBRACKET },
	{ "::=", sizeof ("::=")-1, EQUALS },
	{ "(", sizeof ("(")-1, LEFTPAREN },
	{ ")", sizeof (")")-1, RIGHTPAREN },
	{ ",", sizeof (",")-1, COMMA },
	{ NULL }
};

void tok::init()
{
	tok	*tp;
	char	*cp;
	int	h;
	int	b;

	ZeroMemory(buckets, sizeof(buckets));

	for (tp = tokens; tp->name; tp++) {
		for (h = 0, cp = tp->name; *cp; cp++)
			h += *cp;
		tp->hash = h;
		b = BUCKET(h);
		if (buckets[b])
		    tp->next = buckets[b]; /* BUG ??? ̔˂ */
		buckets[b] = tp;
	}
}

/*
 * Parses a token from the file.  The type of the token parsed is returned,
 * and the text is placed in the string pointed to by token.
 */
int tok::get_token(LPSTR& fp, std::string& token)
{
    static char last = ' ';
    char ch;
    register int hash = 0;
    register struct tok *tp;

	token.erase();
    ch = last;
    /* skip all white space */
    while(isspace(ch) && ch){
		ch = *fp++;
		if (ch == '\n')
			Line++;
    }
    if (!ch) {
		return ENDOFFILE;
    } else if (ch == '"') {
		return parseQuoteString(fp, token);
    }

    /*
     * Accumulate characters until end of token is found.  Then attempt to
     * match this token as a reserved word.  If a match is found, return the
     * type.  Else it is a label.
     */
    do {
		if (ch == '\n')
			Line++;
		if (isspace(ch) || ch == '(' || ch == ')' || ch == '{' || ch == '}' ||
			ch == ',' || ch == ';'){
			if (!isspace(ch) && token.empty()){
				hash += ch;
				token += ch;
				last = ' ';
			} else {
				last = ch;
			}

			for (tp = buckets[BUCKET(hash)]; tp; tp = tp->next) {
				if ((tp->hash == hash) && (token == tp->name))
					break;
			}
			if (tp){
				if (tp->token == CONTINUE)
					continue;
				return (tp->token);
			}

			if (token[0] == '-' && token[1] == '-'){
				/* strip comment */
				if (ch != '\n'){
					while ((ch = *fp++) != 0)
						if (ch == '\n'){
							Line++;
							break;
						}
					}
					if (!ch)
						return ENDOFFILE;
					last = ch;
					return get_token(fp, token);		
			}
			for(std::string::iterator it = token.begin(); it != token.end(); it++)
				if (!isdigit(*it))
					return LABEL;
			return NUMBER;
		} else {
			hash += ch;
			token += ch;
		}
    
	} while ((ch = *fp++) != 0);
    return ENDOFFILE;
}

int tok::parseQuoteString(LPSTR& fp, std::string& token)
{
    char ch = ' ';
	token.erase();

    while(ch) {
        ch = *fp++;
		if (ch == '\n')
			Line++;
		else if (ch == '"') {
            return QUOTESTRING;
        }
    }

    return 0;
}

/*
 * This routine parses a string like  { blah blah blah } and returns OBJID if
 * it is well formed, and NULL if not.
 */
int tok::tossObjectIdentifier(LPSTR& fp)
{
    register int ch = *fp++;

	/*    ch = last; = ' '? */
    /* skip all white space */
    while(ch && isspace(ch)){
        ch = *fp++;
        if (ch == '\n')
            Line++;
    }
    if (ch != '{')
        return 0;

    while(ch) {
        ch = *fp++;

        if (ch == '\n')
            Line++;
        else if (ch == '}')
            return OBJID;
    }

/*    last = ch;*/
    return 0;
}

static void print_error(const char* string, const char* token, int type)
{
	char temp[32];
	std::string s;
    if (type == ENDOFFILE)
	{
		s = string;
		s += "(EOF): On or around line ";
	}
    else if (token)
	{
		s = string;
		s += "(";
		s += token;
		s += "): On or around line ";
	}
    else
	{
		s = string;
		s += ": On or around line ";
	}
	s += itoa(tok::Line, temp, 10);
	throw new std::string(s);
}
////////////////////////////////////////////////////////////////////////////////////////
CMIBNode::CMIBNode()
{
	type = 0;
}

CMIBNode::CMIBNode(const CMIBNode& r)
	: m_label(r.m_label), enums(r.enums), description(r.description)
{
    subid = r.subid;
	m_parent = r.m_parent;
    type = r.type;
}

int CMIBNode::GetHash()
{
	int hash = 0;
	for(std::string::iterator it = m_parent.begin(); it != m_parent.end(); it++)
		hash += *it;
	return hash;
}

void CMIBNode::CreateTree(std::list<CMIBNode>& list, LPSTR& fp)
{
	std::string token;
	std::string name;
    int	type(LABEL);
	enum {
		BETWEEN_MIBS,
		IN_MIB,
	} state  = BETWEEN_MIBS;

	tok::init();

    while(type != ENDOFFILE){
		type = tok::get_token(fp, token);
skipget:
		if (type == END){
		    if (state != IN_MIB){
				print_error("Error, end before start of MIB.", (char *)NULL, type);
				return;
			}
			state = BETWEEN_MIBS;
			continue;
		} else if (type != LABEL){
			if (type == ENDOFFILE){
				return;
			}
			print_error(token.c_str(), "is a reserved word", type);
			return;
		}
		name = token;
		type = tok::get_token(fp, token);
		if (type == DEFINITIONS){
			if (state != BETWEEN_MIBS){
				print_error("Error, nested MIBS.", (char *)NULL, type);
				return;
			}
			state = IN_MIB;
			if (!parse_mib_header(fp, name)){
				print_error("Bad parse of module header", (char *)NULL, type);
				return;
			}
		} else if (type == OBJTYPE){
			parse_objecttype(list, fp, name);
		} else if (type == OBJGROUP){
			parse_objectgroup(list, fp, name);
		} else if (type == NOTIFTYPE){
			parse_notificationDefinition(list, fp, name);
		} else if (type == COMPLIANCE){
			parse_compliance(list, fp, name);
		} else if (type == MODULEIDENTITY){
			parse_moduleIdentity(list, fp, name);
		} else if (type == OBJID){
			parse_objectid(list, fp, name);
		} else if (type == EQUALS){
			if (!parse_asntype(fp, name, type, token)){
				print_error("Bad parse of ASN type definition.", NULL, EQUALS);
				return;
			}
			goto skipget;
		} else if (type == ENDOFFILE){
			break;
		} else {
			print_error("Bad operator", (char *)NULL, type);
			return ;
		}
    }
}

bool CMIBNode::parse_mib_header(LPSTR& fp, std::string& name)
{
    int type = DEFINITIONS;
	std::string token;
    
    /* This probably isn't good enough.  If there is no
       imports clause we can't go around waiting (forever) for a semicolon.
       We need to check for semi following an EXPORTS clause or an IMPORTS
       clause of both.  Look for BEGIN; in my initial MIBs to see those
       that I needed to hack to get to parse because they didn't have
       an IMPORTS or and EXPORTS clause.
       */
    while(type != SEMI){
		type = tok::get_token(fp, token);
    }
    return true;
}

/*
 * Parse an entry of the form:
 * label OBJECT IDENTIFIER ::= { parent 2 }
 * The "label OBJECT IDENTIFIER" portion has already been parsed.
 * Returns 0 on error.
 */
void CMIBNode::parse_objectid(MIBList& list, LPSTR& fp, std::string& name)
{
    register int count;
	std::string token;
    Subid *op, *nop;
    int length;
    Subid oid[32];

    int type = tok::get_token(fp, token);
    if (type != EQUALS){
		print_error("Bad format", token.c_str(), type);
		return;
    }
    if ((length = getoid(fp, oid, 32)) != 0){
		CMIBNode node;
		/*
		* For each parent-child subid pair in the subid array,
		* create a node and link it into the node list.
		*/
		for(count = 0, op = oid, nop=oid+1; count < (length - 2); count++,
			op++, nop++){
			/* every node must have parent's name and child's name or number */
			if (op->second.empty() == false && (nop->second.empty() == false || (nop->first != -1))){
				node.m_parent = op->second;
				if (nop->second.empty() == false)
					node.m_label = nop->second;
				if (nop->first != -1)
					node.subid = nop->first;
				node.type = 0;
				list.push_back(node);
			}
		}
		/*
		* The above loop took care of all but the last pair.  This pair is taken
		* care of here.  The name for this node is taken from the label for this
		* entry.
		* np still points to an unused entry.
		*/
		if (count == (length - 2)){
			if (op->second.empty() == false){
				node.m_parent = op->second;
				node.m_label = name;
				if (nop->first != -1)
					node.subid = nop->first;
				else
					print_error("Warning: This entry is pretty silly", nop->second.c_str(), type);
				list.push_back(node);
			}
		} else {
			print_error("Missing end of oid", (char *)NULL, type);
		}
    } else {
		print_error("Bad object identifier", (char *)NULL, type);
    }
}

/*
 * Parses an asn type.  Structures are ignored by this parser.
 * Returns NULL on error.
 */
bool CMIBNode::parse_asntype(LPSTR& fp, std::string& name, int& ntype, std::string& ntoken)
{
	std::string token;
    EnumList enlist;
    
    int type = tok::get_token(fp, token);
    if (type == SEQUENCE){
		while((type = tok::get_token(fp, token)) != ENDOFFILE){
			if (type == RIGHTBRACKET){
				ntype = tok::get_token(fp, ntoken);
				return true;
			}
		}
		print_error("Expected \"}\"", token.c_str(), type);
		return false;
    } else {
		if (token == "TEXTUAL-CONVENTION"){
			while (type != SYNTAX)
				type = tok::get_token(fp, token);
			type = tok::get_token(fp, token);
		}
		/* textual convention */
		tc* tcp = tc::search_tc(0);
		if (tcp == NULL) {
			print_error("No more textual conventions possible.", token.c_str(), type);
			return false;
		}
		tcp->setDescriptor(name);
		if (!(type & SYNTAX_MASK)){
			print_error("Textual convention doesn't map to real type.", token.c_str(), type);
			return false;
		}
		tcp->setType(type);

		ntype = tok::get_token(fp, ntoken);
		if (ntype == LEFTPAREN){
			int level = 1;
			/* don't record any constraints for now */
			while(level > 0){
				ntype = tok::get_token(fp, ntoken);
				if (ntype == LEFTPAREN)
					level++;
				if (ntype == RIGHTPAREN)
					level--;		
			}
			ntype = tok::get_token(fp, ntoken);
		} else if (ntype == LEFTBRACKET) {
			/* if there is an enumeration list, parse it */
			while((type = tok::get_token(fp, token)) != ENDOFFILE){
				if (type == RIGHTBRACKET)
					break;
				if (type == LABEL){
					/* this is an enumerated label */
					/* a reasonable approximation for the length */
					EnumInts e(0, token);
					type = tok::get_token(fp, token);
					if (type != LEFTPAREN){
						print_error("Expected \"(\"", token.c_str(), type);
						return false;
					}
					type = tok::get_token(fp, token);
					if (type != NUMBER){
						print_error("Expected integer", token.c_str(), type);
						return false;
					}
					e.first = atoi(token.c_str());
					tcp->addEnum(e);
					type = tok::get_token(fp, token);
					if (type != RIGHTPAREN){
						print_error("Expected \")\"", token.c_str(), type);
						return false;
					}
				}
			}
			if (type == ENDOFFILE){
				print_error("Expected \"}\"", token.c_str(), type);
				return false;
			}
			ntype = tok::get_token(fp, ntoken);
		}
		return true;
    }
}


/*
 * Parses an OBJECT TYPE macro.
 */
void CMIBNode::parse_objecttype(MIBList& list, LPSTR& fp, std::string& name)
{
    Subid oid[32];
	std::string token;
	std::string syntax;
	std::string nexttoken;

    int type = tok::get_token(fp, token);
    if (type != SYNTAX){
		print_error("Bad format for OBJECT TYPE", token.c_str(), type);
		return ;
    }
	CMIBNode node;
    type = tok::get_token(fp, token);
    if (type == LABEL){
		type = tc::get_tc(token, node.enums);
    }
    node.type = type;
    int nexttype = tok::get_token(fp, nexttoken);
    switch(type){
	case SEQUENCE:
		syntax = token;
	    if (nexttype == OF){
			syntax += " ";
			syntax += nexttoken;
			nexttype = tok::get_token(fp, nexttoken);
			syntax += " ";
			syntax += nexttoken;
			nexttype = tok::get_token(fp, nexttoken);
	    }
	    break;
	case INTEGER:
	case UINTEGER32:
		syntax = token;
	    if (nexttype == LEFTBRACKET) {
		/* if there is an enumeration list, parse it */
			while((type = tok::get_token(fp, token)) != ENDOFFILE){
				if (type == RIGHTBRACKET)
					break;
				if (type == LABEL){
					/* this is an enumerated label */
					/* a reasonable approximation for the length */
					EnumInts e(0, token);
					node.enums.push_back(e);
					type = tok::get_token(fp, token);
					if (type != LEFTPAREN){
						print_error("Expected \"(\"", token.c_str(), type);
						return;
					}
					type = tok::get_token(fp, token);
					if (type != NUMBER){
						print_error("Expected integer", token.c_str(), type);
						return;
					}
					e.first = atoi(token.c_str());
					node.enums.push_back(e);
					type = tok::get_token(fp, token);
					if (type != RIGHTPAREN){
						print_error("Expected \")\"", token.c_str(), type);
						return;
					}
				}
			}
			if (type == ENDOFFILE){
				print_error("Expected \"}\"", token.c_str(), type);
				return;
			}
			nexttype = tok::get_token(fp, nexttoken);
		} else if (nexttype == LEFTPAREN){
			/* ignore the "constrained integer" for now */
			nexttype = tok::get_token(fp, nexttoken);
			nexttype = tok::get_token(fp, nexttoken);
			nexttype = tok::get_token(fp, nexttoken);
	    }
	    break;
	case BITSTRING:
		syntax = token;
	    if (nexttype == LEFTBRACKET) {
			/* if there is an enumeration list, parse it */
			while((type = tok::get_token(fp, token)) != ENDOFFILE){
				if (type == RIGHTBRACKET)
					break;
				if (type == LABEL){
					/* this is an enumerated label */
					/* a reasonable approximation for the length */
					EnumInts e(0, token);
					type = tok::get_token(fp, token);
					if (type != LEFTPAREN){
						print_error("Expected \"(\"", token.c_str(), type);
						return;
					}
					type = tok::get_token(fp, token);
					if (type != NUMBER){
						print_error("Expected integer", token.c_str(), type);
						return;
					}
					e.first = atoi(token.c_str());
					node.enums.push_back(e);
					type = tok::get_token(fp, token);
					if (type != RIGHTPAREN){
						print_error("Expected \")\"", token.c_str(), type);
						return;
					}
				}
			}
			if (type == ENDOFFILE){
				print_error("Expected \"}\"", token.c_str(), type);
				return;
			}
			nexttype = tok::get_token(fp, nexttoken);
		} else if (nexttype == LEFTPAREN){
			/* ignore the "constrained integer" for now */
			nexttype = tok::get_token(fp, nexttoken);
			nexttype = tok::get_token(fp, nexttoken);
			nexttype = tok::get_token(fp, nexttoken);
	    }
	    break;
	case OCTETSTR:
		syntax = token;
        /* ignore the "constrained octet string" for now */
		if (nexttype == LEFTPAREN) {
			nexttype = tok::get_token(fp, nexttoken);
			if (nexttype == SIZE) {
				nexttype = tok::get_token(fp, nexttoken);
				if (nexttype == LEFTPAREN) {
					nexttype = tok::get_token(fp, nexttoken); /* 0..255 */
					nexttype = tok::get_token(fp, nexttoken); /* ) */
					nexttype = tok::get_token(fp, nexttoken); /* ) */
					if (nexttype == RIGHTPAREN)
					{
						nexttype = tok::get_token(fp, nexttoken);
						break;
					}
				}
			}
			print_error("Bad syntax", token.c_str(), type);
			return;
		}
	    break;
	case OBJID:
	case NETADDR:
	case IPADDR:
	case COUNTER:
	case GAUGE:
	case TIMETICKS:
	case OPAQUE:
	case NUL:
	case LABEL:
	case NSAPADDRESS:
	case COUNTER64:
		syntax = token;
	    break;
	default:
	    print_error("Bad syntax", token.c_str(), type);
	    return;
    }
    if (nexttype == UNITS){
		type = tok::get_token(fp, token);
		if (type != QUOTESTRING) {
			print_error("Bad DESCRIPTION", token.c_str(), type);
			return;
		}
		nexttype = tok::get_token(fp, nexttoken);
    }
    if (nexttype != ACCESS){
		print_error("Should be ACCESS", nexttoken.c_str(), nexttype);
		return;
    }
    type = tok::get_token(fp, token);
    if (type != READONLY && type != READWRITE && type != WRITEONLY
		&& type != NOACCESS && type != READCREATE){
		print_error("Bad access type", nexttoken.c_str(), nexttype);
		return;
    }
    type = tok::get_token(fp, token);
    if (type != STATUS){
		print_error("Should be STATUS", token.c_str(), nexttype);
		return;
    }
    type = tok::get_token(fp, token);
    if (type != MANDATORY && type != CURRENT && type != OPTIONAL && type != OBSOLETE && type != DEPRECATED){
		print_error("Bad status", token.c_str(), type);
		return;
    }
    /*
     * Optional parts of the OBJECT-TYPE macro
     */
    type = tok::get_token(fp, token);
    while (type != EQUALS) {
      switch (type) {
	  case DESCRIPTION:
          type = tok::get_token(fp, token);
          if (type != QUOTESTRING) {
              print_error("Bad DESCRIPTION", token.c_str(), type);
              return;
          }
		  node.description = token;
		  break;
	  case REFERENCE:
		  type = tok::get_token(fp, token);
		  if (type != QUOTESTRING) {
			  print_error("Bad DESCRIPTION", token.c_str(), type);
			  return;
		  }
		  break;
	  case INDEX:
	  case DEFVAL:
	  case AUGMENTS:
	  case NUM_ENTRIES:
          if (tok::tossObjectIdentifier(fp) != OBJID) {
              print_error("Bad Object Identifier", token.c_str(), type);
              return;
          }
          break;

	  default:
          print_error("Bad format of optional clauses", token.c_str(),type);
          return;

      }
      type = tok::get_token(fp, token);
    }
    if (type != EQUALS){
		print_error("Bad format", token.c_str(), type);
		return;
    }
    int length = getoid(fp, oid, 32);
    if (length > 1 && length <= 32){
		/* just take the last pair in the oid list */
		if (oid[length - 2].second.length())
			node.m_parent = oid[length - 2].second;
		node.m_label = name;
		if (oid[length - 1].first != -1)
			node.subid = oid[length - 1].first;
    } else {
		print_error("No end to oid", (char *)NULL, type);
		return;
    }
	list.push_back(node);
}


/*
 * Parses an OBJECT GROUP macro.
 */
void CMIBNode::parse_objectgroup(std::list<CMIBNode>& list, LPSTR& fp, std::string& name)
{
	std::string token;
    Subid oid[32];

    CMIBNode node;
    int type = tok::get_token(fp, token);
    while (type != EQUALS) {
      switch (type) {
      case DESCRIPTION:
          type = tok::get_token(fp, token);
          if (type != QUOTESTRING) {
              print_error("Bad DESCRIPTION", token.c_str(), type);
          }
		  node.description = token;
          break;

	  default:
		  /* NOTHING */
		  break;
      }
      type = tok::get_token(fp, token);
    }
    int length = getoid(fp, oid, 32);
    if (length > 1 && length <= 32){
		/* just take the last pair in the oid list */
		if (oid[length - 2].second.length())
			node.m_parent = oid[length - 2].second;
		node.m_label = name;
		if (oid[length - 1].first != -1)
			node.subid = oid[length - 1].first;
    } else {
		print_error("No end to oid", (char *)NULL, type);
    }
	list.push_back(node);
}

/*
 * Parses a NOTIFICATION-TYPE macro.
 */
void CMIBNode::parse_notificationDefinition(MIBList& list, LPSTR& fp, std::string& name)
{
	std::string token;
    Subid oid[32];

	CMIBNode node;
    int type = tok::get_token(fp, token);
    while (type != EQUALS) {
      switch (type) {
	  case DESCRIPTION:
          type = tok::get_token(fp, token);
          if (type != QUOTESTRING) {
              print_error("Bad DESCRIPTION", token.c_str(), type);
          }
		  node.description = token;
          break;

	  default:
		  /* NOTHING */
		  break;
      }
      type = tok::get_token(fp, token);
    }
    int length = getoid(fp, oid, 32);
    if (length > 1 && length <= 32){
		/* just take the last pair in the oid list */
		if (oid[length - 2].second.length())
			node.m_parent = oid[length - 2].second;
		node.m_label = name;
		if (oid[length - 1].first != -1)
			node.subid = oid[length - 1].first;
    } else {
		print_error("No end to oid", (char *)NULL, type);
	}
	list.push_back(node);
}

/*
 * Parses a compliance macro
 */
void CMIBNode::parse_compliance(MIBList& list, LPSTR& fp, std::string& name)
{
	std::string token;
    Subid oid[32];

	CMIBNode node;
    int type = tok::get_token(fp, token);
    while (type != EQUALS) {
		type = tok::get_token(fp, token);
    }
    int length = getoid(fp, oid, 32);
    if (length > 1 && length <= 32){
	/* just take the last pair in the oid list */
		if (oid[length - 2].second.length())
			node.m_parent = oid[length - 2].second;
		node.m_label = name;
	if (oid[length - 1].first != -1)
	    node.subid = oid[length - 1].first;
    } else {
		print_error("No end to oid", (char *)NULL, type);
    }
	list.push_back(node);
}



/*
 * Parses a module identity macro
 */
void CMIBNode::parse_moduleIdentity(MIBList& list, LPSTR& fp, std::string& name)
{
	std::string token;
    Subid oid[32];

	CMIBNode node;
    int type = tok::get_token(fp, token);
    while (type != EQUALS) {
		type = tok::get_token(fp, token);
    }
    int length = getoid(fp, oid, 32);
    if (length > 1 && length <= 32){
		/* just take the last pair in the oid list */
		if (oid[length - 2].second.length())
			node.m_parent = oid[length - 2].second;
		node.m_label = oid[length - 1].second;
		if (oid[length - 1].first != -1)
			node.subid = oid[length - 1].first;
    } else {
		print_error("No end to oid", (char *)NULL, type);
    }
	list.push_back(node);
}

/*
 * Takes a list of the form:
 * { iso org(3) dod(6) 1 }
 * and creates several nodes, one for each parent-child pair.
 * Returns NULL on error.
 */
int CMIBNode::getoid(LPSTR& fp, Subid* oid,  int length)
{
    register int count;
    int type;
	std::string token;

    if ((type = tok::get_token(fp, token)) != LEFTBRACKET){
		print_error("Expected \"{\"", token.c_str(), type);
    }
    type = tok::get_token(fp, token);
    for(count = 0; count < length; count++){
		oid[count].first = -1;
		if (type == RIGHTBRACKET){
			return count;
		} else if (type != LABEL && type != NUMBER){
			print_error("Not valid for object identifier", token.c_str(), type);
		}
		if (type == LABEL){
			/* this entry has a label */
			oid[count].second = token;
			type = tok::get_token(fp, token);
			if (type == LEFTPAREN){
				type = tok::get_token(fp, token);
				if (type == NUMBER){
					oid[count].first = atoi(token.c_str());
					if ((type = tok::get_token(fp, token)) != RIGHTPAREN){
						print_error("Unexpected a closing parenthesis", token.c_str(), type);
					}
				} else {
					print_error("Expected a number", token.c_str(), type);
				}
			} else {
				continue;
			}
		} else {
			/* this entry  has just an integer sub-identifier */
			oid[count].first = atoi(token.c_str());
		}
		type = tok::get_token(fp, token);
    }
    return count;
}

/////////////////////////////////////////////////////////////////////////////////

CMIB::CMIB()
	: next_peer(NULL), parent(NULL), subid(0), type(0), child_list(NULL)
{
}

CMIB::CMIB(const char* p, int n, CMIB* pParent)
	: child_list(NULL), next_peer(NULL), parent(pParent), label(p), subid(n), type(0)
{
}

CMIB::~CMIB()
{
	delete child_list;
}

void CMIB::DestroyTree(CMIB* pRoot)
{
	delete pRoot;
/*
	CMIB* p = pRoot;
	while (pRoot)
	{
		pRoot = p->next_peer;
		delete p;
	}
*/
}

/*
 * Parse an entry of the form:
 * label OBJECT IDENTIFIER ::= { parent 2 }
 * The "label OBJECT IDENTIFIER" portion has already been parsed.
 */
CMIB* CMIB::CreateTree(LPSTR& fp)
{
	CMIB* tp;
	MIBList nodes;
	CMIBNode::CreateTree(nodes, fp);
	if (nodes.size() == 0)
	{
		MessageBox(NULL, _T("Mib table is bad.  Exiting"), _T("SNMPCtl"), MB_OK);
	}
	else
	{
		build_translation_table();
		/* grow tree from this root node */
		init_node_hash(nodes);
#ifdef _DEBUG
		print_node_hash();
#endif

		/* build root node */
		CMIB* mib = new CMIB("joint-iso-ccitt", 2);

		/* XXX nodes isn't needed in do_subtree() ??? */
		mib->do_subtree();

		/* build root node */
		CMIB* mi2 = new CMIB("ccitt", 0);
		mi2->next_peer = mib;
		/* XXX nodes isn't needed in do_subtree() ??? */
		mi2->do_subtree();

		/* build root node */
		tp = new CMIB("iso", 1);
		tp->next_peer = mi2;
		/* XXX nodes isn't needed in do_subtree() ??? */
		tp->do_subtree();

#ifdef _DEBUG
		tp->print_subtree(0);
#endif /* _DEBUG */

		int bucket;
		bool nodes_left(false);
		/* If any nodes are left, the tree is probably inconsistent */
		for(bucket = 0; bucket < NHASHSIZE; bucket++){
			if (nbuckets[bucket].size()) {
				nodes_left = true;
				break;
			}
		}
		if (nodes_left){
			ATLTRACE(_T("The mib description doesn't seem to be consistent.\n"));
			ATLTRACE(_T("Some nodes couldn't be linked under the \"iso\" tree.\n"));
			ATLTRACE(_T("these nodes are left:\n"));
			for(bucket = 0; bucket < NHASHSIZE; bucket++){
				for(MIBListIter np = nbuckets[bucket].begin(); np != nbuckets[bucket].end(); np++)
					ATLTRACE(_T("%s ::= { %s %ld } (%d)\n"), (*np).GetLabel().c_str(),
					(*np).GetParent().c_str(), (*np).GetSubid(), (*np).GetType());
			}
		}
	}
	return tp;
}

int CMIB::translation_table[256];
bool CMIB::s_tableInitialized(false);
void CMIB::build_translation_table()
{
	if (s_tableInitialized) return;
	s_tableInitialized = true;

    int count;

    for(count = 0; count < 256; count++){
	switch(count){
	    case OBJID:
			translation_table[count] = TYPE_OBJID;
			break;
		case OCTETSTR:
			translation_table[count] = TYPE_OCTETSTR;
			break;
	    case INTEGER:
			translation_table[count] = TYPE_INTEGER;
			break;
	    case NETADDR:
			translation_table[count] = TYPE_IPADDR;
			break;
	    case IPADDR:
			translation_table[count] = TYPE_IPADDR;
			break;
	    case COUNTER:
			translation_table[count] = TYPE_COUNTER;
			break;
		case GAUGE:
			translation_table[count] = TYPE_GAUGE;
			break;
	    case TIMETICKS:
			translation_table[count] = TYPE_TIMETICKS;
			break;
	    case OPAQUE:
			translation_table[count] = TYPE_OPAQUE;
			break;
	    case NUL:
			translation_table[count] = TYPE_NULL;
			break;
	    case COUNTER64:
			translation_table[count] = TYPE_COUNTER64;
			break;
	    case BITSTRING:
			translation_table[count] = TYPE_BITSTRING;
			break;
	    case NSAPADDRESS:
			translation_table[count] = TYPE_NSAPADDRESS;
			break;
	    case UINTEGER32:
			translation_table[count] = TYPE_UINTEGER;
			break;
	    default:
			translation_table[count] = TYPE_OTHER;
			break;
		}
    }
}

MIBList CMIB::nbuckets[NHASHSIZE];

void CMIB::init_node_hash(MIBList& list)
{
	 for (int i = 0; i < NHASHSIZE; i++)
		 nbuckets[i].clear();

	 for (MIBListIter ix = list.begin(); ix != list.end(); ix++)
	 {
         int hash = (*ix).GetHash();
		 nbuckets[NBUCKET(hash)].push_front(*ix);
     }
}

#ifdef _DEBUG
void CMIB::print_node_hash()
{
	for (int i = 0; i < NHASHSIZE; i++)
	{
		if (nbuckets[i].size())
		{
			ATLTRACE(_T("Hash = %d\n"), i);
			for (MIBListIter it = nbuckets[i].begin(); it != nbuckets[i].end(); it++)
				ATLTRACE(_T("   %s\n"), (*it).GetLabel().c_str());
		}
	}
}
#endif

/*
 * Find all the children of root in the list of nodes.  Link them into the
 * tree and out of the nodes list.
 */
void CMIB::do_subtree()
{
	MIBList cchild_list;
    
    int hash = 0;
    for(std::string::iterator cp = label.begin(); cp != label.end(); cp++)
        hash += *cp;
    MIBList& headp = nbuckets[NBUCKET(hash)];
    /*
     * Search each of the nodes for one whose parent is root, and
     * move each into a separate list.
     */
	MIBList oldList;
    for(MIBListIter np = headp.begin(); np != headp.end(); np++) {
		if ((*np).IsParent(label) == false) {
			if ((*np).IsLabel(label)){
				/* if there is another node with the same label, assume that
				* any children after this point in the list belong to the other node.
				* This adds some scoping to the table and allows vendors to
				* reuse names such as "ip".
				*/
				for (; np != headp.end(); np++)
					oldList.push_back(*np);
				break;
			}
			oldList.push_back(*np);
		} else {
			cchild_list.push_front(*np);
			/* take this node out of the node list */
		}
    }
	headp = oldList;
    /*
     * Take each element in the child list and place it into the tree.
     */
	CMIB* peer;
    for(np = cchild_list.begin(); np != cchild_list.end(); np++){
		CMIB* tp = new CMIB((*np).GetLabel().c_str(), (*np).GetSubid(), this);
		tp->parent = this;
		tp->next_peer = NULL;
		tp->type = translation_table[(*np).GetType()];
		tp->enums = (*np).GetEnums();
		tp->description = (*np).GetDescription(); /* steals memory from np */
		if (child_list == NULL)
			child_list = tp;
		else
			peer->next_peer = tp;
		peer = tp;
	    tp->do_subtree();	/* recurse on this child if it isn't
					   an end node */
    }
}

//
// pMibNULL̉\邽߁Astatic֐ƂĎB
//	{́AċAĂяo@ɎĔÓI֐ƂقǂB
CMIB* CMIB::GetSymbol(oid* objid, int oidlen, CMIB* pMib, std::string& buf)
{
	int i;
	CMIB* subtree = pMib;
    for(; subtree; subtree = subtree->next_peer){
		if (*objid == subtree->subid){
			buf += subtree->label;
			goto found;
		}
    }

    /* subtree not found */
	for (i = 0; i < oidlen; i++)
	{	/* output rest of name, uninterpreted */
		char buff[16];
		wsprintfA(buff, "%lu.", *objid++);
		buf += buff;
    }
	buf.erase(buf.size() - 1, 1);
    return NULL;

found:
	CMIB* return_tree = NULL;
    if (oidlen > 1){
		buf += '.';
		return_tree = GetSymbol(objid + 1, oidlen - 1, subtree->child_list, buf);
    } 
    if (return_tree != NULL)
		return return_tree;
    else
		return subtree;
}

void CMIB::ParseSubTree(LPSTR& input, std::vector<oid>& output)
{
	std::string to;
    oid subid = 0;
	CMIB* tp = NULL;
    /*
     * No empty strings.  Can happen if there is a trailing '.' or two '.'s
     * in a row, i.e. "..".
     */
	if ((*input == '\0') ||	(*input == '.'))
		return;

    if (isdigit(*input)) {
		/*
		 * Read the number, then try to find it in the subtree.
		 */
		while (isdigit(*input)) {
			subid *= 10;
			subid += *input++ - '0';
		}
		for (tp = this; tp; tp = tp->next_peer) {
			if (tp->subid == subid)
				goto found;
		}
		tp = NULL;
    }
    else {
		/*
		 * Read the name into a buffer.
		 */
		while ((*input != '\0') &&
	       (*input != '.')) {
			to += *input++;
		}

		/*
		 * Find the name in the subtree;
		 */
		for (tp = this; tp; tp = tp->next_peer) {
			if (lc_cmp(tp->label.c_str(), to.c_str()) == 0) {
				subid = tp->subid;
				goto found;
			}
		}

		/*
		 * If we didn't find the entry, punt...
		 */
		if (tp == NULL) {
			std::string se("sub-identifier not found: ");
			se += to;
			throw new std::string(se);
		}
    }

found:
    if(subid > (u_long)MAX_SUBID){
		std::string se("sub-identifier too large: ");
		se += to;
		throw new std::string(se);
    }

	output.push_back(subid);

    if (*input != '.')
		return;
	if (tp)
		tp->child_list->ParseSubTree(++input, output);
}

int CMIB::lc_cmp(const char* s1, const char* s2)
{
    char c1, c2;

    while(*s1 && *s2){
		if (isupper(*s1))
			c1 = tolower(*s1);
		else
			c1 = *s1;
		if (isupper(*s2))
			c2 = tolower(*s2);
		else
			c2 = *s2;
		if (c1 != c2)
			return ((c1 - c2) > 0 ? 1 : -1);
		s1++;
		s2++;
    }

    if (*s1)
		return -1;
    if (*s2)
		return 1;
    return 0;
}
